#! /usr/bin/env python

"""
"""

# Copyright (C) 2012-2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals

from builtins import str

import hashlib
import os
import re
import paramiko

# EC2 APIs
try:
    import boto
    import boto.ec2.regioninfo
    import boto.exception
except ImportError as err:
    from gc3libs.exceptions import ConfigurationError
    raise ConfigurationError(
        "The EC2 backend is used but the `boto` module"
        " cannot be used: {err}. Please, either install it with"
        "`pip install boto` and verify that it works"
        " by running `python -c 'import boto'`,"
        " then try again, or update your configuration file and"
        " disable any EC2-based resources."
        .format(err=err))

import Crypto.PublicKey.DSA
import Crypto.PublicKey.RSA

# GC3Pie imports
import gc3libs
import gc3libs.defaults
from gc3libs.exceptions import UnrecoverableError, \
    ConfigurationError, MaximumCapacityReached, ResourceNotReady, \
    UnrecoverableAuthError, TransportError
import gc3libs.url
from gc3libs import Run
from gc3libs.utils import same_docstring_as, insert_char_every_n_chars
from gc3libs.backends import LRMS
from gc3libs.backends.shellcmd import ShellcmdLrms
from gc3libs.backends.vmpool import VMPool, InstanceNotFound


## module metadata
__docformat__ = 'reStructuredText'


## module-level constants
available_subresource_types = [gc3libs.defaults.SHELLCMD_LRMS]

# example Boto error message:
# <Response><Errors><Error><Code>TooManyInstances</Code><Message>Quota exceeded for ram: Requested 8000, but already used 16000 of 16384 ram</Message></Error></Errors><RequestID>req-c219213b-88d2-42dc-a3ab-10ac80aa7df7</RequestID></Response>  # noqa
#
_BOTO_ERRMSG_RE = re.compile(r'<Code>(?P<code>[A-Za-z0-9]+)</Code>'
                             '<Message>(?P<message>.*)</Message>', re.X)


## code

class EC2VMPool(VMPool):

    """
    Specific implementation of VMPool
    """

    def _get_instance(self, vm_id):
        # contact EC2 API to get VM info
        try:
            reservations = self.conn.get_all_instances(instance_ids=[vm_id])
        except boto.exception.EC2ResponseError as err:
            # scrape actual error kind and message out of the
            # exception; we do this mostly for sensible logging, but
            # could be an actual improvement to Boto to provide
            # different exception classes based on the <Code>
            # element...
            # XXX: is there a more robust way of doing this?
            match = _BOTO_ERRMSG_RE.search(str(err))
            if match:
                raise UnrecoverableError(
                    "Error getting info on VM %s: EC2ResponseError/%s: %s"
                    % (vm_id, match.group('code'), match.group('message')),
                    do_log=True)
            else:
                # fall back to normal reporting...
                raise UnrecoverableError(
                    "Error getting VM %s: %s" % (vm_id, err),
                    do_log=True)
        if not reservations:
            raise InstanceNotFound(
                "No instance with id %s has been found." % vm_id)

        instances = dict((i.id, i) for i in reservations[0].instances
                         if reservations)
        if vm_id not in instances:
            raise UnrecoverableError(
                "No instance with id %s has been found." % vm_id)

        return instances[vm_id]


class EC2Lrms(LRMS):

    """
    EC2 resource.
    """
    RESOURCE_DIR = '$HOME/.gc3/ec2.d'

    def __init__(self, name,
                 # these parameters are inherited from the `LRMS` class
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime,
                 # these are specific of the EC2Lrms class
                 ec2_region, keypair_name, public_key, vm_auth,
                 image_id=None, image_name=None, ec2_url=None,
                 instance_type=None, auth=None, vm_pool_max_size=None,
                 user_data=None, **extra_args):
        LRMS.__init__(
            self, name,
            architecture, max_cores, max_cores_per_job,
            max_memory_per_core, max_walltime, auth, **extra_args)

        self.free_slots = int(max_cores)
        self.user_run = 0
        self.user_queued = 0
        self.queued = 0
        self.vm_pool_max_size = vm_pool_max_size
        if vm_pool_max_size is not None:
            try:
                self.vm_pool_max_size = int(self.vm_pool_max_size)
            except ValueError:
                raise ConfigurationError(
                    "Value for `vm_pool_max_size` must be an integer,"
                    " was %s instead." % vm_pool_max_size)

        self.subresource_type = self.type.split('+', 1)[1]
        if self.subresource_type not in available_subresource_types:
            raise UnrecoverableError("Invalid resource type: %s" % self.type)

        self.region = ec2_region

        # Mapping of job.execution._lrms_vm_id => LRMS
        self.subresources = {}

        auth = self._auth_fn()
        self.ec2_access_key = auth.ec2_access_key
        self.ec2_secret_key = auth.ec2_secret_key
        if ec2_url is None:
            ec2_url = os.getenv('EC2_URL')
        if ec2_url is None:
            raise gc3libs.exceptions.InvalidArgument(
                "Cannot connect to the EC2 API:"
                " No 'EC2_URL' environment variable defined,"
                " and no 'ec2_url' argument passed to the EC2 backend.")
        self.ec2_url = gc3libs.url.Url(ec2_url)

        # Keypair names can only contain alphanumeric chars!
        if re.match(r'.*\W.*', keypair_name):
            raise ConfigurationError(
                "Keypair name `%s` is invalid: keypair names can only contain "
                "alphanumeric chars: [a-zA-Z0-9_]" % keypair_name)
        self.keypair_name = keypair_name
        self.public_key = os.path.expanduser(
            os.path.expandvars(public_key.strip()))
        self.image_id = image_id
        self.image_name = image_name
        self.instance_type = instance_type
        self._instance_type_specs = {}
        self.user_data = user_data

        self._parse_security_group()
        self._conn = None

        # `self.subresource_args` is used to create subresources
        self.subresource_args = extra_args
        self.subresource_args['type'] = self.subresource_type
        self.subresource_args['architecture'] = self['architecture']
        self.subresource_args['max_cores'] = self['max_cores']
        self.subresource_args['max_cores_per_job'] = self['max_cores_per_job']
        self.subresource_args['max_memory_per_core'] = \
            self['max_memory_per_core']
        self.subresource_args['max_walltime'] = self['max_walltime']
        # SSH-specific configuration
        self.subresource_args['transport'] = 'ssh'
        self.subresource_args['auth'] = vm_auth
        self.subresource_args['ssh_timeout'] = 7  # FIXME: hard-coded!
        self.subresource_args['ignore_ssh_host_keys'] = True
        self.subresource_args['keyfile'] = self.public_key
        if self.subresource_args['keyfile'].endswith('.pub'):
            keyfile = self.subresource_args['keyfile'][:-len('.pub')]
            self.subresource_args['keyfile'] = keyfile
        # ShellcmdLrms by default trusts the configuration, instead of
        # checking the real amount of memory and number of cpus, but
        # we need the real values instead.
        if self.subresource_type == gc3libs.defaults.SHELLCMD_LRMS:
            self.subresource_args['override'] = 'True'

        if not image_name and not image_id:
            raise ConfigurationError(
                "No `image_id` or `image_name` has been specified in the"
                " configuration file.")

    def _connect(self):
        """
        Connect to the EC2 endpoint and check that the required
        `image_id` exists.
        """
        if self._conn is not None:
            return

        args = {
            'aws_access_key_id': self.ec2_access_key,
            'aws_secret_access_key': self.ec2_secret_key,
        }

        if self.ec2_url:
            region = boto.ec2.regioninfo.RegionInfo(
                name=self.region,
                endpoint=self.ec2_url.hostname,
            )
            args['region'] = region
            args['port'] = self.ec2_url.port
            args['host'] = self.ec2_url.hostname
            args['path'] = self.ec2_url.path
            if self.ec2_url.scheme in ['http']:
                args['is_secure'] = False

        self._conn = boto.connect_ec2(**args)
        # Set up the VMPool persistent class. This has been delayed
        # until here because otherwise self._conn is None
        pooldir = os.path.join(os.path.expandvars(EC2Lrms.RESOURCE_DIR),
                               'vmpool', self.name)
        self._vmpool = EC2VMPool(pooldir, self._conn)

        try:
            all_images = self._conn.get_all_images()
        except boto.exception.EC2ResponseError as ex:
            if ex.status == 404:
                # NotFound, probaly the endpoint is wrong
                raise RuntimeError(
                    "Unable to contact the EC2 endpoint at `%s`. Please, "
                    "verify that the URL in the configuration file is "
                    "correct." % (self.ec2_url,))
            raise RuntimeError(
                "Unknown error while connecting to the EC2 endpoint `%s`: "
                "%s" % (self.ec2_url, ex))
        if self.image_id:
            if self.image_id not in [i.id for i in all_images]:
                raise RuntimeError(
                    "Image with id `%s` not found. Please specify a valid"
                    " image name or image id in the configuration file."
                    % self.image_id)
        else:
            images = [i for i in all_images if i.name == self.image_name]
            if not images:
                raise RuntimeError(
                    "Image with name `%s` not found. Please specify a valid"
                    " image name or image id in the configuration file."
                    % self.image_name)
            elif len(images) != 1:
                raise RuntimeError(
                    "Multiple images found with name `%s`: %s"
                    " Please specify an unique image id in configuration"
                    " file." % (self.image_name, [i.id for i in images]))
            self.image_id = images[0].id

    def _create_instance(self, image_id, instance_type=None, user_data=None):
        """
        Create an instance using the image `image_id` and instance
        type `instance_type`. If not `instance_type` is defined, use
        the default.

        This method will also setup the keypair and the security
        groups, if needed.
        """
        self._connect()

        args = {
            'key_name': self.keypair_name,
            'min_count': 1,
            'max_count': 1
        }
        if instance_type:
            args['instance_type'] = instance_type

        if user_data:
            args['user_data'] = user_data

        # Check if the desired keypair is present
        keypairs = dict((k.name, k) for k in self._conn.get_all_key_pairs())
        if self.keypair_name not in keypairs:
            gc3libs.log.info(
                "Keypair `%s` not found: creating it using public key `%s`"
                % (self.keypair_name, self.public_key))
            # Create keypair if it does not exist and give an error if it
            # exists but have different fingerprint
            self._import_keypair()
        else:
            self._have_keypair(keypairs[self.keypair_name])

        # Setup security groups
        if 'security_group_name' in self:
            self._setup_security_groups()
            args['security_groups'] = [self.security_group_name]

        # FIXME: we should add check/creation of proper security
        # groups
        gc3libs.log.debug("Create new VM using image id `%s`", image_id)
        try:
            reservation = self._conn.run_instances(image_id, **args)
        except boto.exception.EC2ResponseError as err:
            # scrape actual error kind and message out of the
            # exception; we do this mostly for sensible logging, but
            # could be an actual improvement to Boto to provide
            # different exception classes based on the <Code>
            # element...
            # XXX: is there a more robust way of doing this?
            match = _BOTO_ERRMSG_RE.search(str(err))
            if match:
                raise UnrecoverableError(
                    "Error starting instance: EC2ResponseError/%s: %s"
                    % (match.group('code'), match.group('message')))
            else:
                # fall back to normal reporting...
                raise UnrecoverableError("Error starting instance: %s" % err)
        except Exception as ex:
            raise UnrecoverableError("Error starting instance: %s" % ex)
        vm = reservation.instances[0]
        self._vmpool.add_vm(vm)
        gc3libs.log.info(
            "VM with id `%s` has been created and is in %s state.",
            vm.id, vm.state)
        return vm

    def _get_subresource(self, vm):
        """
        Return the resource associated to the virtual machine `vm`.

        Updates the internal list of available resources if needed.

        """
        if vm.id not in self.subresources:
            self.subresources[vm.id] = self._make_subresource(
                vm.id, vm.preferred_ip)
        return self.subresources[vm.id]

    def _get_vm(self, vm_id):
        """
        Return the instance with id `vm_id`, raises an error if there
        is no such instance with that id.
        """
        self._connect()
        vm = self._vmpool.get_vm(vm_id)
        return vm

    @staticmethod
    def __pubkey_ssh_fingerprint(privkey):
        """
        Compute SSH public key fingerprint like `ssh-keygen -l -f`.

        We need to convert the key fingerprint from Paramiko's
        internal representation to this colon-separated hex format
        because that's the way the fingerprint is returned from the
        EC2 API.
        """
        return ':'.join((i.encode('hex') for i in privkey.get_fingerprint()))

    @staticmethod
    def __pubkey_aws_fingerprint(privkeypath,
                                 reader=Crypto.PublicKey.RSA.importKey):
        """
        Compute SSH public key fingerprint like AWS EC2 does.

        Return a string representation of the key fingerprint in
        colon-separated hex format (just like OpenSSH commands print
        it to the terminal).

        Argument `privkeypath` is the filesystem path to a file
        containing the private key data.
        """
        with open(privkeypath, 'r') as keydata:
            pubkey = reader(keydata).publickey()
            return insert_char_every_n_chars(
                hashlib.md5(pubkey.exportKey('DER')).hexdigest(), ':', 2)

    def _have_keypair(self, ec2_key):
        """
        Check if the given SSH key is available locally.

        Try to locate the given SSH key first among the keys loaded
        into a running ``ssh-agent`` (if any), then in the key file
        given in the ``public_key`` configuration key.
        """
        # try with SSH agent first
        gc3libs.log.debug("Checking if keypair is registered in SSH agent...")
        agent = paramiko.Agent()
        fingerprints_from_agent = [
            self.__pubkey_ssh_fingerprint(privkey)
            for privkey in agent.get_keys()
        ]
        if ec2_key.fingerprint in fingerprints_from_agent:
            gc3libs.log.debug("Found remote key fingerprint in SSH agent.")
            return True

        # else, try to load from file
        keyfile = self.public_key
        if keyfile.endswith('.pub'):
            keyfile = keyfile[:-4]
        else:
            gc3libs.log.warning(
                "Option `public_key` in configuration file should contain"
                " the path to a public key file (with `.pub` ending),"
                " but `%s` was found instead. Continuing anyway.",
                self.public_key)

        privkey = None
        local_fingerprints = []
        for format, privkey_reader, pubkey_reader in [
                ('DSS', paramiko.DSSKey.from_private_key_file,
                 Crypto.PublicKey.DSA.importKey),
                ('RSA', paramiko.RSAKey.from_private_key_file,
                 Crypto.PublicKey.RSA.importKey),
        ]:
            try:
                gc3libs.log.debug(
                    "Trying to load key file `%s` as SSH %s key...",
                    keyfile, format)
                privkey = privkey_reader(keyfile)
                gc3libs.log.info(
                    "Successfully loaded key file `%s` as SSH %s key.",
                    keyfile, format)
                # compute public key fingerprints, for comparing them
                # with the remote one
                localkey_fingerprints = [
                    # Usual SSH key fingerprint, computed like
                    # `ssh-keygen -l -f` This is used, e.g., by
                    # OpenStack's EC2 compatibility layer.
                    self.__pubkey_ssh_fingerprint(privkey),
                    # Amazon EC2 computes the key fingerprint in a
                    # different way, see http://blog.jbrowne.com/?p=23 and
                    # https://gist.github.com/jtriley/7270594 for details.
                    self.__pubkey_aws_fingerprint(keyfile, pubkey_reader),
                ]
            # PasswordRequiredException < SSHException so we must
            # check this first
            except paramiko.PasswordRequiredException:
                gc3libs.log.warning(
                    "Key %s is encripted with a password, so we cannot check "
                    " if it matches the remote keypair (maybe you should"
                    " start `ssh-agent`?). Continuing without consistency"
                    " check with remote fingerprint.", keyfile)
                return False
            except paramiko.SSHException as ex:
                gc3libs.log.debug(
                    "File `%s` is not a valid %s private key: %s",
                    keyfile, format, ex)
                # try with next format
                continue
        if privkey is None:
            raise ValueError("Public key `%s` is neither a valid"
                             " RSA key nor a DSS key" % self.public_key)

        # check key fingerprint
        if (local_fingerprints
                and ec2_key.fingerprint not in localkey_fingerprints):
            raise UnrecoverableAuthError(
                "Keypair `%s` exists but has different fingerprint than"
                " local one: local public key file `%s` has fingerprint(s)"
                " `%s`, whereas EC2 API reports fingerprint `%s`."
                " Aborting!" % (
                    self.public_key,
                    self.keypair_name,
                    '/'.join(localkey_fingerprints),
                    ec2_key.fingerprint,
                ),
                do_log=True)

    def _import_keypair(self):
        """
        Create a new keypair and import the public key defined in the
        configuration file.
        """
        with open(os.path.expanduser(self.public_key)) as fd:
            try:
                key_material = fd.read()
                imported_key = self._conn.import_key_pair(
                    self.keypair_name, key_material)
                gc3libs.log.info(
                    "Successfully imported key `%s`"
                    " with fingerprint `%s` as keypair `%s`",
                    imported_key.name, imported_key.fingerprint,
                    self.keypair_name)
            except Exception as ex:
                raise UnrecoverableError("Error importing keypair %s: %s"
                                         % (self.keypair_name, ex))

    def _make_subresource(self, id, remote_ip):
        """
        Create a resource associated to the instance with `remote_ip`
        ip using configuration file parameters.
        """
        gc3libs.log.debug(
            "Creating remote ShellcmdLrms resource for IP address %s",
            remote_ip)
        args = self.subresource_args.copy()
        args['frontend'] = remote_ip
        args['name'] = "%s@%s" % (id, self.name)
        return ShellcmdLrms(**args)

    def _parse_security_group(self):
        """
        Parse configuration file and set `self.security_group_rules`
        with a list of dictionaries containing the rule sets
        """
        rules = self.security_group_rules.split(',')
        self.security_group_rules = []
        for rule in rules:
            rulesplit = rule.strip().split(':')
            if len(rulesplit) != 4:
                gc3libs.log.warning("Invalid rule specification in"
                                    " `security_group_rules`: %s" % rule)
                continue
            self.security_group_rules.append({
                'ip_protocol': rulesplit[0],
                'from_port': int(rulesplit[1]),
                'to_port': int(rulesplit[2]),
                'cidr_ip': rulesplit[3],
            })

    def _setup_security_groups(self):
        """
        Check the current configuration and set up the security group
        if it does not exist.
        """
        if not self.security_group_name:
            gc3libs.log.error("Group name in `security_group_name`"
                              " configuration option cannot be empty!")
            return
        security_groups = self._conn.get_all_security_groups()
        groups = dict((g.name, g) for g in security_groups)
        # Check if the security group exists already
        if self.security_group_name not in groups:
            try:
                gc3libs.log.info("Creating security group %s",
                                 self.security_group_name)
                security_group = self._conn.create_security_group(
                    self.security_group_name,
                    "GC3Pie_%s" % self.security_group_name)
            except Exception as ex:
                gc3libs.log.error("Error creating security group %s: %s",
                                  self.security_group_name, ex)
                raise UnrecoverableError(
                    "Error creating security group %s: %s"
                    % (self.security_group_name, ex))

            for rule in self.security_group_rules:
                try:
                    gc3libs.log.debug(
                        "Adding rule %s to security group %s.",
                        rule, self.security_group_name)
                    security_group.authorize(**rule)
                except Exception as ex:
                    if gc3libs.error_ignored(
                            # context:
                            # - module
                            'ec2',
                            # - class
                            'EC2Lrms',
                            # - method
                            'setup_security_groups',
                            # - actual error class
                            ex.__class__.__name__,
                            # - additional keywords
                            'setup',
                            'security',
                            'network',
                            'cloud',
                    ):
                        gc3libs.log.info(
                            "Ignoring error adding rule %s"
                            " to security group %s: %s",
                            rule, self.security_group_name, ex)
                    else:
                        # propagate exception to caller
                        raise
        else:
            # Check if the security group has all the rules we want
            security_group = groups[self.security_group_name]
            current_rules = []
            for rule in security_group.rules:
                rule_dict = {
                    'ip_protocol': rule.ip_protocol,
                    'from_port': int(rule.from_port),
                    'to_port': int(rule.to_port),
                    'cidr_ip': str(rule.grants[0]),
                }
                current_rules.append(rule_dict)

            for new_rule in self.security_group_rules:
                if new_rule not in current_rules:
                    security_group.authorize(**new_rule)

    # Public methods

    def get_image_id_for_job(self, job):
        """
        If a configuration option <application>_image_id is present,
        returns its value, otherwise returns `self.image_id`
        """
        conf_option = job.application_name + '_image_id'
        if conf_option in self:
            return self[conf_option]
        else:
            return self.image_id

    def get_instance_type_for_job(self, job):
        """
        If a configuration option <application>_instance_type is present,
        returns its value, otherwise returns `self.instance_type`
        """
        conf_option = job.application_name + '_instance_type'
        if conf_option in self:
            return self[conf_option]
        else:
            return self.instance_type

    def get_user_data_for_job(self, job):
        """
        If a configuration option <application>_user_data is present,
        returns its value, otherwise returns `self.user_data`.
        """
        conf_option = job.application_name + '_user_data'
        if conf_option in self:
            return self[conf_option]
        else:
            return self.user_data

    @same_docstring_as(LRMS.cancel_job)
    def cancel_job(self, app):
        resource = self._get_subresource(self._get_vm(app.execution._lrms_vm_id))
        return resource.cancel_job(app)

    @same_docstring_as(LRMS.get_resource_status)
    def get_resource_status(self):
        self.updated = False
        # Since we create the resource *before* the VM is actually up
        # & running, it's possible that the `frontend` value of the
        # resources points to a non-existent hostname. Therefore, we
        # have to update them with valid public_ip, if they are
        # present.

        self._connect()
        # Update status of known VMs
        for vm_id in self._vmpool:
            try:
                vm = self._vmpool.get_vm(vm_id)
            except UnrecoverableError as ex:
                gc3libs.log.warning(
                    "Removing stale information on VM `%s`. It has probably"
                    " been deleted from outside GC3Pie.", vm_id)
                self._vmpool.remove_vm(vm_id)
                continue

            vm.update()
            if vm.state == 'pending':
                # If VM is still in pending state, skip creation of
                # the resource
                continue
            elif vm.state == 'error':
                # The VM is in error state: exit.
                gc3libs.log.error(
                    "VM with id `%s` is in ERROR state."
                    " Terminating it!", vm.id)
                vm.terminate()
                self._vmpool.remove_vm(vm.id)
            elif vm.state == 'terminated':
                gc3libs.log.info(
                    "VM `%s` in TERMINATED state. It has probably been"
                    " terminated from outside GC3Pie. Removing it from the"
                    " list of VM.", vm.id)
                self._vmpool.remove_vm(vm.id)
            elif vm.state in ['shutting-down', 'stopped']:
                # The VM has probably ben stopped or shut down from
                # outside GC3Pie.
                gc3libs.log.error(
                    "VM with id `%s` is in terminal state `%s`.",
                    vm.id, vm.state)

            # Get or create a resource associated to the vm
            resource = self._get_subresource(vm)
            try:
                resource.get_resource_status()
            except TransportError as ex:
                for ip in [vm.public_dns_name, vm.private_ip_address]:
                    if vm.preferred_ip == ip:
                        continue
                    vm.preferred_ip = ip
                    resource.frontend = ip
                    gc3libs.log.info(
                        "Connection error. Trying with secondary IP"
                        " address %s", vm.preferred_ip)
                    try:
                        resource.get_resource_status()
                        break
                    except Exception as ex:
                        # XXX: I'm exempting this from the GC3Pie
                        # `error_ignored()` policy, since this is a kind of
                        # "expected" error -- it *will* happen if the VM has
                        # not booted up yet or if we're hunting for the correct
                        # address.
                        gc3libs.log.debug(
                            "Ignoring error in updating resource %s: %s"
                            " The corresponding VM may not be ready yet.",
                            resource.name, ex)
            except Exception as ex:
                # FIXME: Actually, we should try to identify the kind of
                # error we are getting. For instance, if the
                # configuration options `username` is wrong, we will
                # create VMs but we will never be able to submit jobs
                # to them, thus causing an increasing number of
                # useless VMs created on the cloud.
                if gc3libs.error_ignored(
                        # context:
                        # - module
                        'ec2',
                        # - class
                        'EC2Lrms',
                        # - method
                        'get_resource_status',
                        # - actual error class
                        ex.__class__.__name__,
                        # - additional keywords
                        'vm',
                ):
                    gc3libs.log.info(
                        "Ignoring error in updating resource %s: %s"
                        "The corresponding VM may not be ready yet.",
                        resource.name, ex)
                else:
                    # propagate exception to caller
                    raise
            if resource.updated:
                # Update also the instance_type specs, if not
                # already updated
                if not self._instance_type_specs:
                    specs = self._instance_type_specs
                    specs['architecture'] = resource['architecture']
                    specs['max_cores'] = resource['max_cores']
                    specs['max_cores_per_job'] = resource['max_cores_per_job']
                    specs['max_memory_per_core'] = resource['total_memory']
                    self.update(specs)

        self._vmpool.update()
        return self

    @same_docstring_as(LRMS.get_results)
    def get_results(self, app, download_dir, overwrite=False,
                    changed_only=True):
        subresource = self._get_subresource(self._get_vm(app.execution._lrms_vm_id))
        return subresource.get_results(app, download_dir,
                                       overwrite=overwrite,
                                       changed_only=changed_only)

    @same_docstring_as(LRMS.update_job_state)
    def update_job_state(self, app):
        if app.execution._lrms_vm_id not in self.subresources:
            try:
                self.subresources[app.execution._lrms_vm_id] = self._get_subresource(
                    self._get_vm(app.execution._lrms_vm_id))
            except InstanceNotFound as ex:
                gc3libs.log.error(
                    "Changing state of task '%s' to TERMINATED since EC2 "
                    "instance '%s' does not exist anymore.",
                    app.execution.lrms_jobid, app.execution._lrms_vm_id)
                app.execution.state = Run.State.TERMINATED
                raise ex
            except UnrecoverableError as ex:
                gc3libs.log.error(
                    "Changing state of task '%s' to UNKNOWN because of "
                    "an EC2 error.", app.execution.lrms_jobid)
                app.execution.state = Run.State.UNKNOWN
                raise ex

        return self.subresources[app.execution._lrms_vm_id].update_job_state(app)

    def submit_job(self, job):
        """
        Submission on an EC2 resource will usually happen in multiple
        steps, since creating a VM and attaching a resource to it will
        take some time.

        In order to return as soon as possible, the backend will raise
        a `RecoverableError` whenever submission is delayed.

        In case a permanent error is found (for instance, we cannot
        create VMs on the cloud), a `UnrecoverableError` is raised.

        More in detail, when during submission the following will
        happen:

        * First of all, the backend will try to submit the job to one
          of the already available subresources.

        * If none of them is able to sbmit the job, the backend will
          check if there is a VM in pending state, and in case there
          is one it will raise a `RecoverableError`, thus delaying
          submission.

        * If no VM in pending state is found, the `vm_pool_max_size`
          configuration option is checked. If we already reached the
          maximum number of VM, a `UnrecoverableError` is raised.

        * If no VM in pending state is found but `vm_pool_max_size` is
          still lesser than the number of VM currently created (or it
          is None, which for us means no limit), then a new VM is
          created, and `RecoverableError` is raised.

        """
        self._connect()
        # Updating resource is needed to update the subresources. This
        # is not always done before the submit_job because of issue
        # nr.  386:
        #     https://github.com/uzh/gc3pie/issues/386
        self.get_resource_status()

        pending_vms = set(vm.id for vm in self._vmpool.get_all_vms()
                          if vm.state == 'pending')

        image_id = self.get_image_id_for_job(job)
        instance_type = self.get_instance_type_for_job(job)
        # Check that we can actually submit to a flavor like this
        # XXX: this check shouldn't be done by the Engine???
        if self._instance_type_specs:
            specs = self._instance_type_specs
            max_mem = specs['max_memory_per_core']
            max_cpus = specs['max_cores_per_job']
            if (job.requested_memory is not None and
                job.requested_memory > max_mem) \
                or (job.requested_cores is not None and
                    job.requested_cores > max_cpus):
                raise gc3libs.exceptions.LRMSSubmitError(
                    "EC2 flavor %s does not have enough memory/cpus "
                    "to run application %s" % (
                        self.instance_type, job.jobname))

        # First of all, try to submit to one of the subresources.
        for vm_id, resource in list(self.subresources.items()):
            if not resource.updated:
                # The VM is probably still booting, let's skip to the
                # next one and add it to the list of "pending" VMs.
                pending_vms.add(vm_id)
                continue
            try:
                # Check that the required image id and instance type
                # are correct
                vm = self._get_vm(vm_id)
                if (vm.image_id != image_id
                        or vm.instance_type != instance_type):
                    continue
                resource.submit_job(job)
                job.execution._lrms_vm_id = vm_id
                job.changed = True
                gc3libs.log.info(
                    "Job successfully submitted to remote resource %s.",
                    resource.name)
                return job
            except gc3libs.exceptions.LRMSSubmitError as ex:
                if gc3libs.error_ignored(
                        # context:
                        # - module
                        'ec2',
                        # - class
                        'EC2Lrms',
                        # - method
                        'submit_job',
                        # - actual error class
                        ex.__class__.__name__,
                        # - additional keywords
                        'submit',
                ):
                    gc3libs.log.debug(
                        "Ignoring error in submitting to resource %s: %s. ",
                        resource.name, ex)
                else:
                    # propagate exception to caller
                    raise

        # Couldn't submit to any resource.
        if not pending_vms:
            # No pending VM, and no resource available. Create a new VM
            if not self.vm_pool_max_size \
                    or len(self._vmpool) < self.vm_pool_max_size:
                user_data = self.get_user_data_for_job(job)
                vm = self._create_instance(image_id,
                                           instance_type=instance_type,
                                           user_data=user_data)
                pending_vms.add(vm.id)

                self._vmpool.add_vm(vm)
            else:
                raise MaximumCapacityReached(
                    "Already running the maximum number of VM on resource %s:"
                    " %d VMs started, but max %d allowed by configuration."
                    % (self.name, len(self._vmpool), self.vm_pool_max_size),
                    do_log=True)

        # If we reached this point, we are waiting for a VM to be
        # ready, so delay the submission until we wither can submit to
        # one of the available resources or until all the VMs are
        # ready.
        gc3libs.log.debug(
            "No available resource was found, but some VM is still in"
            " `pending` state. Waiting until the next iteration before"
            " creating a new VM. Pending VM ids: %s", pending_vms)
        raise ResourceNotReady(
            "Delaying submission until one of the VMs currently pending"
            " is ready. (Pending VM ids: %r)" % (pending_vms,))

    @same_docstring_as(LRMS.peek)
    def peek(self, app, remote_filename, local_file, offset=0, size=None):
        resource = self._get_subresource(
            self._get_vm(app.execution._lrms_vm_id))
        return resource.peek(app, remote_filename, local_file, offset, size)

    def validate_data(self, data_file_list=None):
        """
        Supported protocols: file
        """
        for url in data_file_list:
            if url.scheme not in ['file']:
                return False
        return True

    def free(self, app):
        """
        Free up any remote resources used for the execution of `app`.
        In particular, this should terminate any remote instance used
        by the job.

        Call this method when `app.execution.state` is anything other
        than `TERMINATED` results in undefined behavior and will
        likely be the cause of errors later on.  Be cautious.
        """
        # XXX: this should probably done only when no other VMs are
        # using this resource.

        # FIXME: freeing the resource from the application is probably
        # not needed since instances are not persistent.

        # freeing the resource from the application is now needed as
        # the same instanc may run multiple applications
        resource = self._get_subresource(self._get_vm(app.execution._lrms_vm_id))
        resource.free(app)

        # FIXME: current approach in terminating running instances:
        # if no more applications are currently running, turn the instance off
        # check with the associated resource
        resource.get_resource_status()
        if not resource.has_running_tasks():
            # turn VM off
            vm = self._get_vm(app.execution._lrms_vm_id)
            gc3libs.log.info("VM instance %s at %s is no longer needed."
                             " Terminating.", vm.id, vm.public_dns_name)
            del self.subresources[vm.id]
            vm.terminate()
            del self._vmpool[vm.id]
            # self._session.save_all()

    @same_docstring_as(LRMS.close)
    def close(self):
        gc3libs.log.info("Closing connection to cloud '%s'...",
                         self.name)
        if self._conn is None and not self.enabled:
            # The resources was most probably disabled by command
            # line. We didn't update it before, so we don't care about
            # currently running VMs now.
            return
        # Update status of VMs and remote resources
        self.get_resource_status()
        for vm_id, resource in list(self.subresources.items()):
            if resource.updated and not resource.has_running_tasks():
                vm = self._get_vm(vm_id)
                gc3libs.log.warning(
                    "VM instance %s at %s is no longer needed. "
                    "You may need to terminate it manually.",
                    vm.id, vm.public_dns_name)
                del self._vmpool[vm.id]
            resource.close()
        # self._session.save_all()


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="ec2",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



# -*- coding: utf-8 -*-

"""
Utility functions for use in unit test code.
"""
#
#  Copyright (C) 2015, 2016, 2018, 2019  University of Zurich. All rights reserved.
#
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or (at your
#  option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

# stdlib imports
from __future__ import absolute_import, print_function, unicode_literals
from builtins import range
from contextlib import contextmanager
from tempfile import NamedTemporaryFile, mkdtemp
import shutil

# GC3Pie imports
from gc3libs import Application, Run
from gc3libs.config import Configuration
from gc3libs.core import Core, Engine
from gc3libs.quantity import GB, MB, hours
from gc3libs.workflow import ParallelTaskCollection, SequentialTaskCollection


@contextmanager
def test_resource(name='test', **params):
    """
    Yield a GC3Pie configuration containing a single resource,
    built using the given parameters.

    The only resource is named ``test`` (can be changed by passing
    keyword argument ``name``).

    .. note::

      The parameters *must* be given in the internal format expected
      by the backend "LRMS" constructors, not in the string format
      expected by the configuration file parser.
    """
    cfg = Configuration()
    cfg.TYPE_CONSTRUCTOR_MAP['noop'] = ('gc3libs.backends.noop', 'NoOpLrms')
    rsc = cfg.resources[name]
    # defaults (1) -- general
    rsc.update(
        name=name,
        type='noop',
        transport='local',
        auth='none',
        architecture=set([Run.Arch.X86_64]),
        enabled=True,
        # Use unusual values so that we can easily spot if the `override` option works
        large_file_chunk_size=1.78 * MB,
        large_file_threshold=1.414 * GB,
        max_cores=123,
        max_cores_per_job=123,
        max_memory_per_core=999 * GB,
        max_walltime=7 * hours,
    )
    # update with given parameters
    rsc.update(**params)
    # defaults (2) -- type-specific
    if rsc.type == 'shellcmd':
        rsc.setdefault('override', False)
        rsc.setdefault('time_cmd', '/usr/bin/time')

    yield cfg

    # since TYPE_CONSTRUCTOR_MAP is a class-level variable, we
    # need to clean up otherwise other tests will see the No-Op
    # backend
    del cfg.TYPE_CONSTRUCTOR_MAP['noop']


@contextmanager
def temporary_core(transition_graph=None, **params):
    with test_resource(**params) as cfg:
        name = params.get('name', 'test')
        core = Core(cfg)
        rsc = core.get_backend(name)
        # give each job a 50% chance of moving from one state to the
        # next one
        if transition_graph:
            rsc.transition_graph = transition_graph

        yield core


@contextmanager
def temporary_directory(*args, **kwargs):
    tmpdir = mkdtemp(*args, **kwargs)
    yield tmpdir
    shutil.rmtree(tmpdir, ignore_errors=True)


@contextmanager
def temporary_engine(transition_graph=None, **kw):
    with temporary_core(transition_graph, **kw) as core:
        yield Engine(core)


@contextmanager
def temporary_config_file(cfgtext=None, keep=False):
    """
    Write a GC3Pie configuration into a temporary file.

    Yields an open file object pointing to the configuration file.  Its
    ``.name`` attribute holds the file path in the filesystem.
    """
    if cfgtext is None:
        cfgtext = ("""
[resource/test]
enabled = yes
type = shellcmd
frontend = localhost
transport = local
max_cores_per_job = 4
max_memory_per_core = 8GiB
max_walltime = 8 hours
max_cores = 10
architecture = x86_64
auth = none
override = no
            """)
    with NamedTemporaryFile(prefix='gc3libs.test.',
                            suffix='.tmp',
                            mode='w+t',
                            delete=(not keep)) as cfgfile:
        cfgfile.write(cfgtext)
        cfgfile.flush()
        yield cfgfile
        # file is automatically deleted upon exit


@contextmanager
def temporary_config(cfgtext=None):
    """
    Return a GC3Pie ``Configuration`` object.

    Optional argument `cfgtext` holds the contents of the configuration file to
    use. If not given, a default one will be used.

    """
    with temporary_config_file(cfgtext) as cfgfile:
        yield Configuration(cfgfile.name)


class SuccessfulApp(Application):
    """An application instance reporting always a zero exit code."""
    def __init__(self, name='success', **extra_args):
        Application.__init__(
            self,
            ['/bin/true'],
            inputs=[],
            outputs=[],
            output_dir='/tmp',
            jobname=name,
            requested_cores=1,
            **extra_args)

    def terminated(self):
        self.execution.returncode = 0


class UnsuccessfulApp(Application):
    """An application reporting always a non-zero exit code."""
    def __init__(self, name='fail', **extra_args):
        Application.__init__(
            self,
            ['/bin/false'],
            inputs=[],
            outputs=[],
            output_dir='/tmp',
            jobname=name,
            requested_cores=1,
            **extra_args)

    def terminated(self):
        self.execution.returncode = (0, 1)


class SimpleParallelTaskCollection(ParallelTaskCollection):
    def __init__(self, num_tasks, **extra_args):
        tasks = [SuccessfulApp('stage{n}'.format(n=n)) for n in range(num_tasks)]
        ParallelTaskCollection.__init__(self, tasks, **extra_args)


class SimpleSequentialTaskCollection(SequentialTaskCollection):
    def __init__(self, num_tasks, **extra_args):
        tasks = [SuccessfulApp('stage{n}'.format(n=n)) for n in range(num_tasks)]
        SequentialTaskCollection.__init__(self, tasks, **extra_args)



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2012, 2015, 2018  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import zip
from builtins import str
from builtins import range
from builtins import object
__docformat__ = 'reStructuredText'

# stdlib imports
import csv
import os
import shutil
import tempfile

import pytest

import sqlalchemy
import sqlalchemy.sql as sql

# GC3Pie imports
import gc3libs.exceptions
from gc3libs.persistence import Persistable, make_store
import gc3libs.persistence.sql
from gc3libs.session import Session
from gc3libs.utils import Struct
from gc3libs import Task
from gc3libs.workflow import TaskCollection


class _PStruct(Struct, Persistable):

    """
    A Persistable+Struct mix-in.

    This class just exists so that we are able to persist something in
    the tests and make a non-trivial equality check afterwards.
    """
    pass


def test_create():
    tmpdir = tempfile.mktemp(
        prefix=(os.path.basename(__file__) + '.'),
        suffix='.d')
    try:
        sess = Session(tmpdir)
        assert os.path.isdir(sess.path)
        sess.destroy()
    except:
        if os.path.exists(tmpdir):
            shutil.rmtree(tmpdir)
        raise


def test_destroy():
    tmpdir = tempfile.mktemp(
        prefix=(os.path.basename(__file__) + '.'),
        suffix='.d')
    try:
        sess = Session(tmpdir)
        tid = sess.add(_PStruct(a=1, b='foo'))
        sess.destroy()
        # destroy should kill all traces of the sessiondir
        assert not os.path.exists(sess.path)
        # in particular, no task can be loaded
        with pytest.raises(gc3libs.exceptions.LoadError):
            sess.load(tid)
    except:
        if os.path.exists(tmpdir):
            shutil.rmtree(tmpdir)
        raise


class TestSession(object):

    @pytest.fixture(autouse=True)
    def _with_tmp_session(self):
        tmpdir = tempfile.mktemp(
            prefix=(os.path.basename(__file__) + '.'),
            suffix='.d')
        self.tmpdir = tmpdir
        self.sess = Session(tmpdir)
        self.extra_args = {}

        yield

        self.sess.destroy()

    def test_session_directory_created(self):
        assert os.path.isdir(self.sess.path)
        assert os.path.samefile(
            os.path.join(os.path.abspath('.'), self.tmpdir),
            self.sess.path)

    def test_store_url_file_exists(self):
        self.sess.flush()
        storefile = os.path.join(self.sess.path, Session.STORE_URL_FILENAME)
        assert os.path.isfile(storefile)

    def test_add(self):
        tid = self.sess.add(_PStruct(a=1, b='foo'))
        assert len(self.sess) == 1
        assert [tid] == self.sess.list_ids()

    def test_add_updates_metadata(self):
        """Check that on-disk metadata is changed on add(..., flush=True)."""
        self.sess.add(_PStruct(a=1, b='foo'), flush=True)
        fd_job_ids = open(os.path.join(self.sess.path,
                                       self.sess.INDEX_FILENAME), 'r')
        ids = fd_job_ids.read().split()
        assert len(ids) == 1
        assert ids == [str(i) for i in self.sess.tasks]

    def test_empty_lines_in_index_file(self):
        """
        Check that the index file is read correctly even when there
        are empty lines.

        Because of a bug, in some cases the
        `gc3libs.session.Session`:class: used to create invalid job
        ids (equal to ``''``).
        """
        self.sess.add(_PStruct(a=1, b='foo'), flush=True)
        with open(os.path.join(self.sess.path,
                               self.sess.INDEX_FILENAME), 'a') as fp:
            fp.write('\n\n\n')
            fp.flush()
        self.sess = Session(self.sess.path, **self.extra_args)
        ids = self.sess.list_ids()
        assert len(ids) == 1
        assert ids == [str(task_id) for task_id in self.sess.tasks]

    def test_add_no_flush(self):
        """Check that metadata is not changed on add(..., flush=False)."""
        tid = self.sess.add(_PStruct(a=1, b='foo'), flush=False)
        # in-memory metadata is updated
        assert len(self.sess) == 1
        assert [tid] == self.sess.list_ids()
        # on-disk metadata is not
        fd_job_ids = open(os.path.join(self.sess.path,
                                       self.sess.INDEX_FILENAME), 'r')
        assert '' == fd_job_ids.read()

    def test_remove(self):
        # add tasks
        tid1 = self.sess.add(_PStruct(a=1, b='foo'))
        tid2 = self.sess.add(_PStruct(a=1, b='foo'))
        assert len(self.sess) == 2
        self.sess.remove(tid1)
        assert len(self.sess) == 1
        self.sess.remove(tid2)
        assert len(self.sess) == 0

    def test_remove_children(self):
        """
        Test if the session is able to remove all children of a task
        """
        # Removing objects
        obj = _PStruct(name='GC3')
        obj.tasks = [_PStruct(name='GC3')]
        id = self.sess.add(obj)
        self.sess.remove(id)

        assert len(self.sess.store.list()) == 0

    def test_reload_session(self):
        self.sess.add(_PStruct(a=1, b='foo'))
        self.sess.add(_PStruct(a=2, b='bar'))
        self.sess.add(_PStruct(a=3, b='baz'))
        sess2 = Session(self.sess.path, **self.extra_args)
        assert len(sess2) == 3
        for task_id in sess2.tasks.keys():
            task = sess2.store.load(task_id)
            assert task == sess2.tasks[task_id]
        for task2_id, task1_id in zip(sorted(sess2.tasks.keys()),
                                      sorted(self.sess.tasks.keys())):
            assert (self.sess.tasks[task1_id] ==
                         sess2.tasks[task2_id])

    def test_incomplete_session_dir(self):
        tmpdir = tempfile.mktemp(
            prefix=(os.path.basename(__file__) + '.'),
            suffix='.d')
        os.mkdir(tmpdir)
        incomplete_sess = Session(tmpdir)
        assert os.path.exists(os.path.join(tmpdir,
                                                Session.INDEX_FILENAME))
        assert os.path.exists(os.path.join(tmpdir,
                                                Session.STORE_URL_FILENAME))
        incomplete_sess.destroy()

    def test_load_external_jobid(self):
        """
        Check if we are able to load an object not belonging to the session.
        """
        obj1 = _PStruct(a=1, b='foo')
        extraid = self.sess.store.save(obj1)
        # note: `load(..., add=True) will also add `obj2` to the
        # session so it gets removed when the session is destroyed.
        obj2 = self.sess.load(extraid, add=True)
        assert obj1 == obj2

    def test_creation_of_timestamp_files(self):
        start_file = os.path.join(self.sess.path,
                                  self.sess.TIMESTAMP_FILES['start'])
        end_file = os.path.join(self.sess.path,
                                self.sess.TIMESTAMP_FILES['end'])

        assert os.path.exists(start_file)
        assert not os.path.exists(end_file)

        assert os.stat(start_file).st_mtime == self.sess.created

        self.sess.set_end_timestamp()
        assert os.path.exists(end_file)
        assert os.stat(end_file).st_mtime == self.sess.finished

    def test_load_session_reads_session_start_time(self):
        """Check if session reads the creation time from the `created` file"""
        session2 = Session(self.sess.path)
        start_file = os.path.join(self.sess.path,
                                  self.sess.TIMESTAMP_FILES['start'])
        assert os.stat(start_file).st_mtime == self.sess.created
        assert os.stat(start_file).st_mtime == session2.created

    def test_standard_session_iterator_for_tasks(self):
        self.sess.add(Task(jobname='task-1'))
        self.sess.add(Task(jobname='task-2'))
        self.sess.add(Task(jobname='task-3'))

        assert (set(('task-1', 'task-2', 'task-3')) ==
                     set(job.jobname for job in self.sess))

    def test_standard_session_iterator_for_tasks_and_task_collections(self):
        coll = TaskCollection(jobname='collection',
                              tasks=[Task() for i in range(3)])
        self.sess.add(coll)

        assert (['collection'] ==
                     [job.jobname for job in self.sess])

    def test_workflow_iterator_for_session(self):
        coll = TaskCollection(
            jobname='collection',
            tasks=[Task(jobname='task-%d' % i) for i in range(3)])
        coll2 = TaskCollection(
            jobname='collection-1',
            tasks=[Task(jobname='task-1-%d' % i) for i in range(3)])
        coll.tasks.append(coll2)

        self.sess.add(coll)

        assert (['collection', 'task-0', 'task-1', 'task-2',
                      'collection-1', 'task-1-0', 'task-1-1', 'task-1-2'] ==
                     [job.jobname for job in self.sess.iter_workflow()])


class StubForSqlSession(TestSession):

    def test_sql_store(self):
        jobid = self.sess.save(_PStruct(a=1, b='foo'))
        self.sess.flush()
        q = (sql.select([self.sess.store.t_store.c.id])
             .where(self.sess.store.t_store.c.id == jobid))
        conn = self.sess.store._engine.connect()
        results = conn.execute(q)
        rows = results.fetchall()
        assert len(rows) == 1
        assert rows[0][0] == jobid

        # remove object from the store, since self.sess.destroy() will
        # not remove it!
        self.sess.store.remove(jobid)


class TestSqliteSession(StubForSqlSession):

    @classmethod
    def setup_class(cls):
        # skip SQLite tests if no SQLite module present (Py 2.4)
        sqlite3 = pytest.importorskip("sqlite3")

    @pytest.fixture(autouse=True)
    def _with_tmp_session(self):
        self.tmpdir = os.path.abspath(
            tempfile.mktemp(
                prefix=(os.path.basename(__file__) + '.'),
                suffix='.d'))
        self.sess = Session(
            self.tmpdir,
            create=True,
            store_or_url="sqlite:///{0}/store.db".format(self.tmpdir))
        self.extra_args = {}

        yield

        if os.path.exists(self.tmpdir):
            shutil.rmtree(self.tmpdir)


class TestMysqlSession(StubForSqlSession):

    @classmethod
    def setup_class(cls):
        # we skip MySQL tests if no MySQLdb module is present
        MySQLdb = pytest.importorskip("MySQLdb")

    @pytest.fixture(autouse=True)
    def _with_tmp_session(self):
        self.tmpdir = tempfile.mktemp(
            prefix=(os.path.basename(__file__) + '.'),
            suffix='.d')
        self.tablename = os.path.basename(tmpdir)
        self.extra_args = {'table_name': self.tablename}
        try:
            self.sess = Session(
                tmpdir,
                store_url="mysql://gc3user:gc3pwd@localhost/gc3",
                **self.extra_args)
        except sqlalchemy.exc.OperationalError:
            if os.path.exists(tmpdir):
                shutil.rmtree(tmpdir)
            pytest.mark.skip("Cannot connect to MySQL database.")

        yield

        self.sess.destroy()
        conn = self.sess.store._engine.connect()
        conn.execute("drop table `%s`" % self.tablename)

# main: run tests

if "__main__" == __name__:
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
"""
Driver script for running the `forwardPremium` application on SMSCG.
"""
# Copyright (C) 2011  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__author__ = 'Benjamin Jonen <benjamin.jonen@bf.uzh.ch'
# summary of user-visible changes
__changelog__ = """

"""
__docformat__ = 'reStructuredText'

# Call: 
# -x /home/benjamin/workspace/idrisk/model/bin/idRiskOut -b /home/benjamin/workspace/idrisk/model/base para.loop  -C 1 -N


# std module imports
from __future__ import absolute_import, print_function
import numpy as np
import os
import sys
import shutil
import copy
import scipy.optimize

#scipy.optimize.fmin_powell
#scipy.optimize.fmin

# set some ugly paths
# export PYTHONPATH=$PYTHONPATH:/home/benjamin/workspace/idrisk/model/code
sys.path.append('/home/benjamin/workspace/idrisk/model/code')

# import personal libraries
path2Pymods = os.path.join(os.path.dirname(__file__), '../')
if not sys.path.count(path2Pymods):
    sys.path.append(path2Pymods)
path2Src = os.path.join(os.path.dirname(__file__), '../src')
if not sys.path.count(path2Src):
    sys.path.append(path2Src)

# Remove all files in curPath if -N option specified. 
if __name__ == '__main__':    
    import sys
    if '-N' in sys.argv:
        path2Pymods = os.path.join(os.path.dirname(__file__), '../')
        if not sys.path.count(path2Pymods):
            sys.path.append(path2Pymods)
        curPath = os.getcwd()
        filesAndFolder = os.listdir(curPath)
        if 'gc3IdRisk.csv' in filesAndFolder or 'idRiskParaSearch.csv' in filesAndFolder: # if another paraSearch was run in here before, clean up. 
            if 'para.loop' in os.listdir(os.getcwd()):
                shutil.copyfile(os.path.join(curPath, 'para.loop'), os.path.join('/tmp', 'para.loop'))
                shutil.rmtree(curPath)
                shutil.copyfile(os.path.join('/tmp', 'para.loop'), os.path.join(curPath, 'para.loop'))
                os.remove(os.path.join('/tmp', 'para.loop'))
            else:
                shutil.rmtree(curPath)
# Call: 
# -x /home/benjamin/workspace/idrisk/model/bin/idRiskOut -b /home/benjamin/workspace/idrisk/model/base para.loop -xVars 'wBarLower' -xVarsDom '-0.2 0.2 ' -target_fx '-0.1' -convCrit '4.9e-2 -sv info  -C 10 -N



# ugly workaround for Issue 95,
# see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import idRiskParaSearch


# superclasses
import forwardPremium

# personal libraries
import costlyOptimization
from createTable import createOverviewTable

# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript
from gc3libs.workflow import SequentialTaskCollection, ParallelTaskCollection
import gc3libs.utils

import gc3libs.debug

# logger
from pymods.support.support import wrapLogger
import pymods.support.support as support
from pymods.classes.tableDict import tableDict

from makePlots import momentPlots

# Set up logger
logger = wrapLogger(loggerName = 'idRiskParaSearchLogger', streamVerb = 'DEBUG', logFile = os.path.join(os.getcwd(), 'idRiskParaSearch.log'))


class solveParaCombination(SequentialTaskCollection):

    def __init__(self, substs, solverParas, **sessionParas):

        logger.debug('entering solveParaCombination.__init__ for job %s' % sessionParas['jobname'])
        self.iter    = 0

        self.jobname = sessionParas['jobname']
        self.substs = substs

        self.sessionParas     = sessionParas
        self.pathToExecutable = sessionParas['pathToExecutable']
        self.architecture     = sessionParas['architecture']
        self.localBaseDir     = sessionParas['localBaseDir']
        
        self.paraFolder = os.path.join(os.getcwd(), sessionParas['jobname'])
        
        self.wBarLower_task = idRiskParaSearchDriver(self.paraFolder, self.substs, solverParas, **sessionParas)

        SequentialTaskCollection.__init__(self, self.jobname, [self.wBarLower_task])

        logger.debug('done gParaSearchDriver.__init__ for job %s' % sessionParas['jobname'])

    def __str__(self):
        return self.jobname

    def next(self, *args): 
        self.iter += 1
        if self.wBarLower_task.execution.returncode == 13:
            logger.critical('wBarLower failed. terminating para combo')
            self.execution.returncode = 13
            return Run.State.TERMINATED
        elif self.wBarLower_task.execution.returncode == 0:
            wBarTable = tableDict.fromTextFile(os.path.join(self.paraFolder, 'optimwBarLower', 'overviewSimu'), width = 20, prec = 10)
            optimalRunTable = wBarTable.getSubset( np.abs(wBarTable['wBarLower'] - self.wBarLower_task.costlyOptimizer.best_x) < 1.e-7 )
            optimalRunFile = open(os.path.join(self.paraFolder, 'optimalRun'), 'w')
            print >> optimalRunFile, optimalRunTable
            return Run.State.TERMINATED
        else:
            print 'unknown return code'
            os._exit
##        if self.beta_task.execution.returncode == 13:
##            logger.critical('beta failed. terminating para combo')
##            self.execution.returncode = 13
##            self.failed = True
##            return Run.State.TERMINATED
##        logger.debug('entering solveParaCombination.next in iteration %s' % self.iter)
##        self.changed = True
##        betaVal = self.beta_task.costlyOptimizer.best_x
##        self.substs['input/parameters.in'].append(('beta', betaVal, '(0,)', 'bar-separated'))
##        if self.iter == 1:
##            # then loop over wbar
##            self.extra_args['jobname'] = self.extra_args['jobname'] + '_' + 'beta' + '=' + str(self.beta_task.costlyOptimizer.best_x)
##            xVar = 'wBarLower'
##            xInitialGuess = [-0.15, 0.15]
##            targetVar = 'iBar_Shock0Agent0'
##            solverParas = {}
##            solverParas['plotting'] = False
##            solverParas['target_fx'] = -0.1
##            solverParas['convCrit'] = 1.e-4
##            self.wBarLower_task = idRiskParaSearchDriver(xVar, xInitialGuess, targetVar, self.paraFolder, self.pathToExecutable, self.architecture, self.localBaseDir, self.substs, solverParas, **self.extra_args)
##            self.add(self.wBarLower_task)
##        else:
##            if self.wBarLower_task.execution.returncode == 13:
##                logger.critical('beta failed. terminating para combo')
##                self.execution.returncode = 13
##                self.failed = True
##                return Run.State.TERMINATED
##            logger.debug('converged for beta and wBarLower for job %s' % self.extra_args['jobname'])
##            wBarTable = tableDict.fromTextFile(os.path.join(self.paraFolder, 'optimwBarLower', 'overviewSimu'), width = 20, prec = 10)
##            optimalRunTable = wBarTable.getSubset( np.abs(wBarTable['wBarLower'] - self.wBarLower_task.costlyOptimizer.best_x) < 1.e-7 )
##            optimalRunFile = open(os.path.join(self.paraFolder, 'optimalRun'), 'w')
##            print >> optimalRunFile, optimalRunTable
##            return Run.State.TERMINATED
##        return Run.State.RUNNING


class idRiskParaSearchDriver(SequentialTaskCollection):
    '''
      Script that executes optimization/solving. Each self.next is one iteration in the solver. 
    '''

    def __init__(self, paraFolder, substs, solverParas, **sessionParas):

        logger.debug('entering gParaSearchDriver.__init__')

        self.jobname = sessionParas['jobname'] + 'driver'
        solverParas['jobname'] = self.jobname
        
        self.sessionParas     = sessionParas
        self.solverParas      = solverParas
        self.substs           = substs
 
        self.paraFolder       = paraFolder
        self.iter             = 0

        # create a subfolder for optim over xVar
        self.optimFolder = os.path.join(paraFolder, 'optim' + str(solverParas['xVars'][0]))
        gc3libs.utils.mkdir(self.optimFolder)

        self.costlyOptimizer = costlyOptimization.costlyOptimization(solverParas)
        
        self.xVars       = solverParas['xVars']
        self.xParaCombos = solverParas['xInitialParaCombo']
        self.targetVars   = solverParas['targetVars']
        self.target_fx   = solverParas['target_fx']

        self.evaluator = idRiskParaSearchParallel(self.xVars, self.xParaCombos, substs, self.optimFolder, solverParas , **sessionParas)

        initial_task = self.evaluator

        SequentialTaskCollection.__init__(self, self.jobname, [initial_task])

        logger.debug('done solveParaCombination.__init__')

    def __str__(self):
        return self.jobname

    def next(self, *args): 
        self.iter += 1
        logger.debug('entering idRiskParaSearchDriver.next in iteration %s for variables %s and paraCombo %s' % (self.iter, self.solverParas['xVars'], self.jobname))
        # sometimes next is called even though run state is terminated. In this case simply return. 
        if self.costlyOptimizer.converged:
            return Run.State.TERMINATED
        self.changed = True
        if  self.execution.state == 'TERMINATED':
            logger.debug('idRiskParaSearchDriver.next already terminated. Returning.. ')
            return Run.State.TERMINATED
        newVals = self.evaluator.target(self.xVars, self.xParaCombos, self.targetVars, self.target_fx)
        if newVals is None:
            logger.critical('evaluating variable %s at guess %s failed' % (self.xVars, self.xParaCombos))	
            self.execution.returncode = 13
            self.failed = True
            return Run.State.TERMINATED
        self.costlyOptimizer.updateInterpolationPoints(self.xParaCombos, newVals)
        if not self.costlyOptimizer.checkConvergence():
            logger.debug('not converged yet. Building new guess and reevaluating. ')
            self.costlyOptimizer.updateApproximation()
            self.xParaCombos = self.costlyOptimizer.generateNewGuess()
            self.evaluator = idRiskParaSearchParallel(self.xVars, self.xParaCombos, self.substs, self.optimFolder, self.solverParas, **self.sessionParas)
            self.add(self.evaluator)
        else:
            logger.debug('converged idRiskParaSearchDriver.next in iteration %s for variables %s. Returning exit code 0. ' % (self.iter, self.solverParas['xVars']))
            self.execution.returncode = 0
            return Run.State.TERMINATED
        logger.debug('done idRiskParaSearchDriver.next in iteration %s for variables %s' % (self.iter, self.solverParas['xVars']))
        return Run.State.RUNNING




class idRiskParaSearchParallel(ParallelTaskCollection, forwardPremium.paraLoop_fp):
    '''
      When initialized generates a number of Applications that are distributed on the grid. 
    '''

    def __str__(self):
        return self.jobname


    def __init__(self, xVars, paraCombos, substs, optimFolder, solverParas, **sessionParas):

        logger.debug('entering idRiskParaSearchParallel.__init__')
        # create jobname
        self.jobname = 'evalSolverGuess' + '_' + sessionParas['jobname']
        for paraCombo in paraCombos:
            self.jobname += str(paraCombo)
            
        self.substs       = substs
        self.optimFolder  = optimFolder
        self.solverParas  = solverParas
        self.sessionParas = sessionParas
        forwardPremium.paraLoop_fp.__init__(self, verbosity = 'INFO')
        tasks = self.generateTaskList(xVars, paraCombos, substs, sessionParas)
        ParallelTaskCollection.__init__(self, self.jobname, tasks)

        logger.debug('done idRiskParaSearchParallel.__init__')

    def target(self, xVars, xParaCombos, targetVars, target_fx):
        '''
          Method that builds an overview table for the jobs that were run and then returns the values. 
        '''
        logger.debug('entering idRiskParaSearchParallel.target. Computing target for xVar = %s, xVals = %s' % (xVars, xParaCombos))
        # Each line in the resulting table (overviewSimu) represents one paraCombo
##        if xVar == 'wBarLower' and xVals[0] == -0.10085714:
##            print 'here'
        overviewTable = createOverviewTable(resultDir = self.optimFolder, outFile = 'simulation.out', exportFileName = 'overviewSimu', sortCols = [], orderCols = [], verb = 'INFO')
        if overviewTable == None:
            logger.critical('overviewTable empty')
            return None
#            os._exit(1)
        xVars = copy.deepcopy(xVars)
        for ixVar, xVar in enumerate(xVars):
            if xVar == 'beta':
                xVars[ixVar] = 'beta_disc'
            else:
                xVars[ixVar] = xVar
        # print table
        #overviewTable.order(['dy'] + xVars)
        #overviewTable.sort(['dy'] + xVars)
        logger.info('table for job: %s' % self.jobname)
        logger.info(overviewTable) 
        # Could replace this with a check that every xVar value is in the table, then output the two relevant columns.
        result = np.array([])
        for xParaCombo in xParaCombos:
            overviewTableSub = copy.deepcopy(overviewTable)        
            for xVar, xVal in zip(xVars, xParaCombo):
                overviewTableSub = overviewTableSub.getSubset( np.abs( overviewTableSub[xVar] - xVal ) < 1.e-8 )
            if len(overviewTableSub) == 0:
                logger.critical('Cannot find value for xVal %s, i.e. overviewTableSub empty. Did you set the pythonPath?' % xVal)
                return None
            elif len(overviewTableSub) == 1:
                #result.append(np.linalg.norm(np.array([ overviewTableSub[targetVars[ixVar]][0] for ixVar, var in enumerate(xVars) ]) - target_fx))
                result = np.append(result, [ overviewTableSub[targetVars[ixVar]][0] for ixVar, var in enumerate(xVars) ])
            else:
                logger.critical('Cannot find unique value for xVal %s' % xVal)
                os._exit(1)
        logger.info('Computed target: Returning result to solver\n')
        return result
        #return result


    def print_status(self, mins,means,vector,txt):
        print txt,mins, means, list(vector)

    def generateTaskList(self, xVars, paraCombos, substs, sessionParas):
        pathToExecutable = sessionParas['pathToExecutable']
        localBaseDir     = sessionParas['localBaseDir']
        architecture     = sessionParas['architecture']
        jobname          = sessionParas['jobname']
        # Fill the task list
        tasks = []
        for paraCombo in paraCombos:
            executable = os.path.basename(pathToExecutable)
            inputs = { pathToExecutable:executable }        
            # make a "stage" directory where input files are collected
            path_to_stage_dir = os.path.join(self.optimFolder, jobname)
            
            # 2. apply substitutions to parameter files in local base dir
            for (path, changes) in substs.iteritems():
                for (var, val, index, regex) in changes:
                    support.update_parameter_in_file(os.path.join(localBaseDir, path), var, index, val, regex)
                # adjust xVar in parameter file
                index = 0
                regex = 'bar-separated'
                for xVar, xVal in zip(xVars, paraCombo):
                    support.update_parameter_in_file(os.path.join(localBaseDir, path), xVar, 0, xVal, regex)
                    path_to_stage_dir += '_' + xVar + '=' + str(xVal)
            gc3libs.utils.mkdir(path_to_stage_dir)
            prefix_len = len(path_to_stage_dir) + 1 
            # fill stage dir
            self.fillInputDir(localBaseDir, path_to_stage_dir)
            # 3. build input file list
            for dirpath,dirnames,filenames in os.walk(path_to_stage_dir):
                for filename in filenames:
                    # cut the leading part, which is == to path_to_stage_dir
                    relpath = dirpath[prefix_len:]
                    # ignore output directory contents in resubmission
                    if relpath.startswith('output'):
                        continue
                    remote_path = os.path.join(relpath, filename)
                    inputs[os.path.join(dirpath, filename)] = remote_path
            # all contents of the `output` directory are to be fetched
            outputs = { 'output/':'' }
            #kwargs = sessionParas.copy()
            kwargs = {}
            kwargs['jobname'] = self.jobname
##            kwargs.pop('requested_memory')
##            kwargs.pop('requested_walltime')
##            kwargs.pop('requested_cores')
            kwargs['stdout'] = 'forwardPremiumOut.log'
            kwargs['join'] = True
            kwargs['output_dir'] = os.path.join(path_to_stage_dir, 'output')
            kwargs['requested_architecture'] = architecture

            # hand over job to create
            tasks.append(forwardPremium.GPremiumApplication('./' + executable, [], inputs, outputs, **kwargs)) 
        return tasks


class idRiskParaSearchScript(SessionBasedScript, forwardPremium.paraLoop_fp):
    """
      Read `.loop` files and execute the `forwardPremium` program accordingly.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = '0.2',
            # only '.loop' files are considered as valid input
            input_filename_pattern = '*.loop',
        )
        forwardPremium.paraLoop_fp.__init__(self, verbosity = 'INFO')

    def setup_options(self):
        self.add_param("-b", "--initial", metavar="DIR",
                       dest="initial",
                       help="Include directory contents in any job's input."
                       " Use this to specify the initial guess files.")
        self.add_param("-n", "--dry-run",
                       dest = 'dryrun', action="store_true", default = False,
                       help="Take the loop for a test run")
        self.add_param("-x", "--executable", metavar="PATH",
                       dest="executable", default=os.path.join(
                           os.getcwd(), "forwardPremiumOut"),
                       help="Path to the `forwardPremium` executable binary"
                       "(Default: %(default)s)")
        self.add_param("-X", "--architecture", metavar="ARCH",
                       dest="architecture", default=Run.Arch.X86_64,
                       help="Processor architecture required by the executable"
                       " (one of: 'i686' or 'x86_64', without quotes)")
        self.add_param("-mP", "--makePlots", metavar="ARCH", type = bool, 
                       dest="makePlots", default = True,
                       help="Generate population plots each iteration.  ")
        self.add_param("-xVars", "--xVars", metavar="ARCH",
                       dest="xVars", default = 'EA',
                       help="x variables over which to optimize")
        self.add_param("-xVarsDom", "--xVarsDom", metavar="ARCH",
                       dest="xVarsDom", default = '0.5 0.9',
                       help="Domain to sample x values from. Space separated list. ")
        self.add_param("-targetVars", "--target_fx", metavar="ARCH",
                       dest="targetVars", default = '0.1',
                       help="Domain to sample x values from. Space separated list. ")
        self.add_param("-target_fx", "--target_fx", metavar="ARCH",
                       dest="target_fx", default = '0.1',
                       help="Domain to sample x values from. Space separated list. ")
        self.add_param("-sv", "--solverVerb", metavar="ARCH",
                       dest="solverVerb", default = 'DEBUG',
                       help="Separate verbosity level for the global optimizer ")
        self.add_param("-yC", "--yConvCrit", metavar="ARCH", type = float, 
                       dest="convCrit", default = '1.e-2',
                       help="Convergence criteria for y variables. ")
#-x /home/benjamin/workspace/idrisk/model/bin/idRiskOut -b /home/benjamin/workspace/idrisk/model/base para.loop -xVars 'wBarLower' -xVarsDom '-0.2 0.2 ' -target_fx '-0.1' -yC '4.9e-2' -sv info  -C 10 -N

    def parse_args(self):
        """
        Check validity and consistency of command-line options.
        """
        if not os.path.exists(self.params.executable):
            raise gc3libs.exceptions.InvalidUsage(
                "Path '%s' to the 'forwardPremium' executable does not exist;"
                " use the '-x' option to specify a valid one."
                % self.params.executable)
        if os.path.isdir(self.params.executable):
            self.params.executable = os.path.join(self.params.executable,
                                                  'forwardPremium')
        gc3libs.utils.check_file_access(self.params.executable, os.R_OK|os.X_OK,
                                gc3libs.exceptions.InvalidUsage)



    def new_tasks(self, extra):     

        paraLoopFile = self._search_for_input_files(self.params.args).pop()  # search_... returns set.

        # Copy base dir
        localBaseDir = os.path.join(os.getcwd(), 'localBaseDir')
        shutil.copytree(self.params.initial, localBaseDir)

        for jobname, substs in self.process_para_file(paraLoopFile):
            # yield job
            sessionParas = {}
            sessionParas['pathToExecutable'] = self.params.executable
            sessionParas['architecture'] = self.params.architecture
            sessionParas['localBaseDir'] = localBaseDir
            sessionParas['jobname'] = jobname
            # Compute domain
            xVarsDom = self.params.xVarsDom.split()
            lowerBds = np.array([xVarsDom[i] for i in range(len(xVarsDom)) if i % 2 == 0], dtype = 'float64')
            upperBds = np.array([xVarsDom[i] for i in range(len(xVarsDom)) if i % 2 == 1], dtype = 'float64')
            domain = zip(lowerBds, upperBds)
            solverParas = {}
            solverParas['xVars'] = self.params.xVars.split()
            solverParas['xInitialParaCombo'] = np.array([lowerBds, upperBds])
            solverParas['targetVars'] = self.params.xVars.split()
            solverParas['target_fx'] = map(float, self.params.target_fx.split())
            solverParas['plotting'] = False
            solverParas['convCrit'] = self.params.convCrit
            yield (jobname, solveParaCombination, [ substs, solverParas ], sessionParas)
            
def extractLinspace(strIn):
    import re
    if re.match('\s*linspace', strIn):
        (linSpacePart, strIn) = re.match('(\s*linspace\(.*?\)\s*)[,\s*]*(.*)', strIn).groups()
        args = re.match('linspace\(([(0-9\.\s-]+),([0-9\.\s-]+),([0-9\.\s-]+)\)', linSpacePart).groups()
        args = [ float(arg) for arg in args] # assume we always want float for linspace
        linSpaceVec = np.linspace(args[0], args[1], args[2])
        return linSpaceVec
    else: 
        print 'cannot find linspace in string'
        os._exit()


if __name__ == '__main__':
    logger.info('Starting: \n%s' % ' '.join(sys.argv))
    idRiskParaSearchScript().run()
    logger.debug('combine resulting tables')    
    tableList = [ os.path.join(os.getcwd(), folder, 'optimalRun') for folder in os.listdir(os.getcwd()) if os.path.isdir(folder) and not folder == 'localBaseDir' and not folder == 'idRiskParaSearch.jobs' ]
    tableDicts = [ tableDict.fromTextFile(table, width = 20, prec = 10) for table in tableList if os.path.isfile(table)]
    if tableDicts:
        optimalRuns = tableDicts[0]
        for ixTable, table in enumerate(tableDicts):
            if ixTable == 0: continue
            optimalRuns = optimalRuns.getAppended(table)
        #optimalRuns.order(['dy', 'wBarLower'])
        #optimalRuns.sort(['dy'])
        logger.info(optimalRuns)
        f = open(os.path.join(os.getcwd(), 'optimalRuns'), 'w')  
        print >> f, optimalRuns
        f.flush()
    #logger.info('Generating plot')
    #baseName = 'moments'
    #path = os.getcwd()
    # conditions = {}
    # overlay = {'EP': True, 'e_rs': True, 'e_rb': True, 'sigma_rs': True, 'sigma_rb': True}
    # tableFile = os.path.join(os.getcwd(), 'optimalRuns')
    # figureFile = os.path.join(os.getcwd(), 'optimalRunsPlot.eps')
    # momentPlots(baseName = baseName, path = path, xVar = 'dy', overlay = overlay, conditions = conditions, tableFile = tableFile, figureFile = figureFile)
    logger.info('main done')


logger.info('done')



#! /usr/bin/env python
#
#   ggamess.py -- Front-end script for submitting multiple GAMESS jobs to SMSCG.
#
#   Copyright (C) 2010-2012  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple GAMESS jobs to SMSCG.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``ggamess --help`` for program usage instructions.
"""
# summary of user-visible changes
__changelog__ = """
  2017-11-09:
    * Initial release.
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__docformat__ = 'reStructuredText'


# stdlib imports
import os
import sys

# GC3Pie imports
import gc3libs
from gc3libs import Application
from gc3libs.cmdline import SessionBasedScript, existing_file
from gc3libs.quantity import days

# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == '__main__':
    import gtrain
    gtrain.GtrainScript().run()


class GtrainScript(SessionBasedScript):
    """
Given a list of directories, each containing a `train.conllu`, 
a `parse.conllu`, and a word embedding file `vectors.xz`, 
run a parser training job for each, possibly in parallel.

The `gtrain` command keeps a record of jobs (submitted, executed and
pending) in a session file (set name with the '-s' option); at each
invocation of the command, the status of all recorded jobs is updated,
output from finished jobs is collected, and a summary table of all
known jobs is printed.  New jobs are added to the session if new input
files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; `gtrain` will delay submission
of newly-created jobs so that this limit is never exceeded.
    """

    def setup_args(self):
        self.add_param("input_dirs", nargs='+',
                       type=existing_file, default=None,
                       help="Directories with input files.")

    def setup_options(self):
        self.add_param('--config', '-c', default='gtrain.cfg',
                       type=existing_file, help="Configuration file.")
        
    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version='1.0',
        )

    def new_tasks(self, extra):
        # create tasks
        for path in self.params.input_dirs:
	    if path.endswith('/'):
		path = path[:-1]
            if not os.path.isdir(path):
                self.log.warning("Path `%s` is no directory! Skipping.", path)
                continue
            train_file = parse_file = embeddings = None
            for entry in os.listdir(path):
                if entry.endswith('train.conllu'):
                    train_file = os.path.join(path, entry)
                elif entry.endswith('dev.conllu'):
                    parse_file = os.path.join(path, entry)
                elif entry.endswith('vectors.xz'):
                    embeddings = os.path.join(path, entry)
                else:
                    continue
            if train_file is None:
                self.log.warning("Path `%s` does not contain a training file! Ignoring.", path)
            if parse_file is None:
                self.log.warning("Path `%s` does not contain a parsing file! Ignoring.", path)
            if embeddings is None:
                self.log.warning("Path `%s` does not contain a word embedding file! Ignoring.", path)
            # create task(s)
            kw = extra.copy()
            kw['jobname'] = os.path.basename(path)
            kw['inputs'] = {
                train_file: 'train.conllu',
                parse_file: 'parse.conllu',
                embeddings: 'vectors.xz',
                self.params.config: 'train.cfg',
            }
            kw['outputs'] = gc3libs.ANY_OUTPUT
            kw['requested_cores'] = 8
            kw['requested_walltime'] = 7*days
            kw['stdout'] = kw['jobname'] + '.log'
            kw['stderr'] = kw['jobname'] + '.log'
            yield Application(
                ['sh', '-c',
                 '. /home/ubuntu/parser/bin/activate'
                 ' &&'
                 ' mkdir -v config && ln -v /home/ubuntu/parser/src/config/defaults.cfg ./config'
                 ' &&'
                 ' sed -re "s:^train_files = .+:train_files = $PWD/train.conllu:" -i train.cfg'
                 ' &&'
                 ' sed -re "s:^parse_files = .+:parse_files = $PWD/parse.conllu:" -i train.cfg'
                 ' &&'
                 ' sed -re "s:^filename = .+:filename = $PWD/vectors.xz:" -i train.cfg'
                 ' &&'
                 ' python /home/ubuntu/parser/src/main.py --save_dir ./output train --config_file train.cfg'],
                **kw
            )



#! /usr/bin/env python
#
#   gscr.py -- Front-end script for running ParRecoveryFun Matlab
#   function with a given combination of reference models.
#
#   Copyright (C) 2014, 2015  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gscr.py --help`` for program usage
instructions.

Input parameters consists of:
@param_folder
@data_folder
@sound_index
...

Options:
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2015-05-04:
  * Added -D and -L option
  -L [INT], --linkrange [INT]
                        Linkrange value. Default: 1.
  -D [INT], --displacement [INT]
                        Displacement value. Default: 10.

  2014-11-13:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gvideotrack
    gvideotrack.GVideoTrackingScript().run()

import os
import sys
import time
import tempfile
import re

import shutil
import random

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

TRAJECTORY_FILE_EXTENSION = ".ijout.txt"

## custom application class
class GVideoTrackingApplication(Application):
    """
    Custom class to wrap the execution of the Matlab script.
    """
    application_name = 'video_track'

    def __init__(self, video_file, **extra_args):

        self.output_dir = extra_args['output_dir']
        # self.result_dir = extra_args['result_dir']

        inputs = dict()
        outputs = dict()
        executables = []

        # Check if binary to be executed is provided as part of input arguments
        if 'R_master' in extra_args:
            inputs[os.path.abspath(extra_args["R_master"])] = "ParticleLinker.R"

        if 'jarfile' in extra_args:
            inputs[os.path.abspath(extra_args["jarfile"])] = "ParticleLinker.jar"

        if 'requested_memory' in extra_args:
            memory = int(extra_args['requested_memory'].amount(Memory.MB))
        else:
            gc3libs.log.warning("Requested memory not set. Using default 1MB")
            memory = 1000

        # Rscript --vanilla s3it_articleLinker.R trj_out/data00422.ijout.txt jar/ParticleLinker.jar result
        # arguments = "Rscript --vanilla ParticleLinker.R %s ParticleLinker.jar result %s" % (os.path.basename(video_file),memory)
        arguments = "Rscript --vanilla ParticleLinker.R %s ParticleLinker.jar result %s %s %s" % (
            os.path.basename(video_file),
            extra_args['linkrange'],
            extra_args['displacement'],
            memory
        )

        inputs[video_file] = os.path.basename(video_file)

        # Set output
        outputs['result/'] = 'result/'


        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gstrj.log',
            join=True,
            executables = executables,
            **extra_args)


class GVideoTrackingScript(SessionBasedScript):
    """
    For each param file (with '.mat' extension) found in the 'param folder',
    GscrScript extracts the corresponding index (from filename) and searches for
    the associated file in 'data folder'. For each pair ('param_file','data_file'),
    GscrScript generates execution Tasks.

    The ``gscr`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gscr`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GVideoTrackingApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GVideoTrackingApplication,
            )

    def setup_args(self):

        self.add_param('videos', type=str,
                       help="Path to the video trajectory files files.")

    def setup_options(self):
        self.add_param("-R", "--Rscript", metavar="[STRING]",
                       dest="R_master", default=None,
                       help="Location of the R script that implements the "
                       "'link_particles' function.")

        self.add_param("-j", "--jarfile", type=str, metavar="[STRING]",
                       dest="jarfile", default=None,
                       help="Location of the 'ParticleLinker.jar'.")

        self.add_param("-L", "--linkrange", type=int, metavar="[INT]",
                       dest="linkrange", default=1,
                       help="Linkrange value. Default: 1.")

        self.add_param("-D", "--displacement", type=int, metavar="[INT]",
                       dest="displacement", default=10,
                       help="Displacement value. Default: 10.")

    def parse_args(self):
        """
        Check validity of input parameters and selected benchmark.
        """

        if not os.path.isdir(self.params.videos):
            raise OSError("No such file or directory: %s ",
                          os.path.abspath(self.params.videos))

        if self.params.R_master:
            if not os.path.isfile(self.params.R_master):
                raise gc3libs.exceptions.InvalidUsage("link_particle function "
                                                      " file %s not found"
                                                      % self.params.R_master)

        if self.params.jarfile:
            if not os.path.isfile(self.params.jarfile):
                raise gc3libs.exceptions.InvalidUsage("ParticleLinker jar "
                                                      " file %s not found"
                                                      % self.params.jarfile)


        assert int(self.params.linkrange)
        assert int(self.params.displacement)

    def new_tasks(self, extra):
        """
        For each of the network data and for each of the selected benchmarks,
        create a GscrApplication.

        First loop the input files, then loop the selected benchmarks
        """
        tasks = []

        for video_file in os.listdir(self.params.videos):

            if not video_file.endswith(TRAJECTORY_FILE_EXTENSION):
                gc3libs.log.info("Ingoring input file %s. "
                                 "Not compliant with expected file extension."
                                 % video_file)
                continue

            # Extract filename
            jobname = video_file.split(TRAJECTORY_FILE_EXTENSION)[0].strip()

            extra_args = extra.copy()

            if self.params.R_master:
                extra_args['R_master'] = self.params.R_master

            if self.params.jarfile:
                extra_args['jarfile'] = self.params.jarfile

            extra_args['linkrange'] = self.params.linkrange
            extra_args['displacement'] = self.params.displacement

            extra_args['jobname'] = jobname

            tasks.append(GVideoTrackingApplication(
                os.path.join(self.params.videos,video_file),
                **extra_args))

        return tasks



#! /usr/bin/env python
#
#   gbenchmark.py -- Front-end script for benchmarking statistical
#   Business software on a common set of network data.
#
#   Copyright (C) 2011, 2012  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for running selected benchmark statistical business
software.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gbenchmark.py --help`` for program usage
instructions.

Input parameters consists of:
:param str network map files: Path to folder containing all network files
(in .dat format) in the for of:

...

Options:
:benchmark str benchmark suite: String identifying the benchmark suite to be
used. Available benchmarks are:
 . Infomap

Important: All benchmark have to be executed on the same hardware in order to
provide consistent benchmark results.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2014-03-14:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@gc3.uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gbenchmarkallinone
    gbenchmarkallinone.GbenchmarkScript().run()

import os
import sys
import time
import tempfile

import shutil
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

BENCHMARKS=['infomap', 'lprop', 'leadingeigenvector', 'fastgreedy', 'multilevel', 'walktrap']
DEFAULT_BENCHMARKS = "infomap, lprop, leadingeigenvector, fastgreedy, multilevel, walktrap"

"""
* Run all benchmarks for each network file on the same VM
* There will be more network files
* Put all network files in single folder
* Run all benchmakrs in sequence
* Make sure each benchmark is executed on an exclusive core
* Run 6 VMs (2 cores each: 1 for OS 1 for R execution)
  * Each VM runs all network files against single benchmark
* Space not an issur: copy all selected network files in the destination VM
* Copy results back
* Use same image file with iGraph already installed
* Use same flavor as before (to guarantee that each VM runs of separate hardware)
* Results are collected only at the end when all network files have been processed by a ginven algorithm
"""

## custom application class
class GbenchmarkApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'benchmark'

    def __init__(self, network_data_path, benchmark_name, benchmark_file, **extra_args):

        self.output_dir = extra_args['output_dir']

        inputs = dict()

        inputs[network_data_path] = './network_data/'
        inputs[benchmark_file] = os.path.basename(benchmark_file)

        # adding wrapper main script
        gbenchmark_wrapper = resource_filename(Requirement.parse("gc3pie"),
                                                  "gc3libs/etc/gbenchmark_wrapper_allinone.py")

        inputs[gbenchmark_wrapper] = "gbenchmark_wrapper.py"

        arguments = "python gbenchmark_wrapper.py %s %s " % (inputs[benchmark_file], inputs[network_data_path])

        # Take full node from 'benchmark' flavor
        extra_args['requested_cores'] = 8

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = ["./results"],
            stdout = 'gbenchmark.log',
            join=True,
            executables = ['gbenchmark_wrapper.py', inputs[benchmark_file]],
            **extra_args)

class GbenchmarkScript(SessionBasedScript):
    """
    Fro each network file (with '.dat' extension) found in the 'input folder',
    GbenchmarkScript generates as many Tasks as 'benchmarks' defined.

    The ``gbenchmark`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gbenchmark`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GbenchmarkApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GbenchmarkApplication,
            )

    def setup_options(self):
        self.add_param("-b", "--benchmark", metavar="[STRING]",
                       dest="benchmarks", default=DEFAULT_BENCHMARKS,
                       help="Comma separated list of benchmarks that " \
                       " should be executed. " \
                       "Default: %s." % DEFAULT_BENCHMARKS)

        self.add_param("-D", "--benchmark_scripts_location", metavar="[STRING]",
                       dest="benchmarks_location", default=None,
                       help="Execution scripts for the benchmarks")

    def setup_args(self):

        self.add_param('network_path', type=str,
                       help="Path to the folder containing network files.")

    def parse_args(self):
        """
        Check validity of input parameters and selected benchmark.
        """

        # check args:
        if not os.path.isdir(self.params.network_path):
            raise gc3libs.exceptions.InvalidUsage(
                "Invalid path to network data: '%s'."
                % self.params.network_path)

        if not os.path.isdir(self.params.benchmarks_location):
            raise gc3libs.exceptions.InvalidUsage(
                "Invalid path to benchmarks scripts location: '%s'."
                % self.params.benchmarks_location)

        # Clear benchmarks list
        self.params.benchmarks = [ bench.strip() for bench in self.params.benchmarks.split(",") ]

        self.benchmarks = dict()

        for test in os.listdir(self.params.benchmarks_location):
            benchmark_name = test.strip().split(".")[0]
            if benchmark_name in self.params.benchmarks:
                gc3libs.log.info("Adding benchmark [%s] with execution file [%s]" % (benchmark_name,
                                                                                     os.path.join(self.params.benchmarks_location,
                                                                                                  test)))
                self.benchmarks[benchmark_name] = os.path.join(self.params.benchmarks_location,test)
                self.params.benchmarks.remove(benchmark_name)

    def new_tasks(self, extra):
        """
        For each of the network data and for each of the selected benchmarks,
        create a GbenchmarkApplication.

        First loop the input files, then loop the selected benchmarks
        """
        tasks = []

        for (benchmark_name,benchmark_file) in self.benchmarks.items():

            # XXX: need to find a more compact name
            jobname = "gbenchmark-%s" % (benchmark_name)

            extra_args = extra.copy()

            extra_args['jobname'] = jobname

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', benchmark_name)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', benchmark_name)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', benchmark_name)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', benchmark_name)

            tasks.append(GbenchmarkApplication(
                os.path.abspath(self.params.network_path),
                benchmark_name,
                benchmark_file,
                **extra_args))

        return tasks



# Original source from http://cci.lbl.gov/cctbx_sources/scitbx/differential_evolution.py
# broader web site http://www.icsi.berkeley.edu/~storn/code.html#pyth

'''
*** Copyright Notice ***

cctbx Copyright (c) 2006, The Regents of the University of
California, through Lawrence Berkeley National Laboratory (subject to
receipt of any required approvals from the U.S. Dept. of Energy).  All
rights reserved.

If you have questions about your rights to use or distribute this
software, please contact Berkeley Lab's Technology Transfer Department
at  TTD@lbl.gov referring to "cctbx (LBNL Ref CR-1726)"

This software package includes code written by others which may be
governed by separate license agreements.  Please refer to the associated
licenses for further details.

NOTICE.  This software was developed under funding from the U.S.
Department of Energy.  As such, the U.S. Government has been granted for
itself and others acting on its behalf a paid-up, nonexclusive,
irrevocable, worldwide license in the Software to reproduce, prepare
derivative works, and perform publicly and display publicly.  Beginning
five (5) years after the date permission to assert copyright is obtained
from the U.S. Department of Energy, and subject to any subsequent five
(5) year renewals, the U.S. Government is granted for itself and others
acting on its behalf a paid-up, nonexclusive, irrevocable, worldwide
license in the Software to reproduce, prepare derivative works,
distribute copies to the public, perform publicly and display publicly,
and to permit others to do so.
'''

#from scitbx.array_family import flex
#from stdlib import random
from __future__ import absolute_import, print_function
import numpy as np
class differential_evolution_optimizer(object):
  """
This is a python implementation of differential evolution
It assumes an evaluator class is passed in that has the following
functionality
data members:
  n              :: The number of parameters
  domain         :: a  list [(low,high)]*n
                   with approximate upper and lower limits for each parameter
  x              :: a place holder for a final solution

  also a function called 'target' is needed.
  This function should take a parameter vector as input and return a the function to be minimized.

  The code below was implemented on the basis of the following sources of information:
  1. http://www.icsi.berkeley.edu/~storn/code.html
  2. http://www.daimi.au.dk/~krink/fec05/articles/JV_ComparativeStudy_CEC04.pdf
  3. http://ocw.mit.edu/NR/rdonlyres/Sloan-School-of-Management/15-099Fall2003/A40397B9-E8FB-4B45-A41B-D1F69218901F/0/ses2_storn_price.pdf


  The developers of the differential evolution method have this advice:
  (taken from ref. 1)

If you are going to optimize your own objective function with DE, you may try the
following classical settings for the input file first: Choose method e.g. DE/rand/1/bin,
set the number of parents NP to 10 times the number of parameters, select weighting
factor F=0.8, and crossover constant CR=0.9. It has been found recently that selecting
F from the interval [0.5, 1.0] randomly for each generation or for each difference
vector, a technique called dither, improves convergence behaviour significantly,
especially for noisy objective functions. It has also been found that setting CR to a
low value, e.g. CR=0.2 helps optimizing separable functions since it fosters the search
along the coordinate axes. On the contrary this choice is not effective if parameter
dependence is encountered, something which is frequently occuring in real-world optimization
problems rather than artificial test functions. So for parameter dependence the choice of
CR=0.9 is more appropriate. Another interesting empirical finding is that rasing NP above,
say, 40 does not substantially improve the convergence, independent of the number of
parameters. It is worthwhile to experiment with these suggestions. Make sure that you
initialize your parameter vectors by exploiting their full numerical range, i.e. if a
parameter is allowed to exhibit values in the range [-100, 100] it's a good idea to pick
the initial values from this range instead of unnecessarily restricting diversity.

Keep in mind that different problems often require different settings for NP, F and CR
(have a look into the different papers to get a feeling for the settings). If you still
get misconvergence you might want to try a different method. We mostly use DE/rand/1/... or DE/best/1/... .
The crossover method is not so important although Ken Price claims that binomial is never
worse than exponential. In case of misconvergence also check your choice of objective
function. There might be a better one to describe your problem. Any knowledge that you
have about the problem should be worked into the objective function. A good objective
function can make all the difference.

Note: NP is called population size in the routine below.)
Note: [0.5,1.0] dither is the default behavior unless f is set to a value other then None.

  """

  def __init__(self,
               evaluator,
               population_size=50,
               f=None,   # de stepsize
               cr=0.9,   # looks like this is the inverse of the crossover prob, i.e staying prob.
               eps=1e-2,
               n_cross=1,
               max_iter=10000,
               monitor_cycle=200,
               out=None,
               show_progress=False,
               show_progress_nth_cycle=1,
               insert_solution_vector=None,
               dither_constant=0.4):
    
    # Set running variables
    self.dither=dither_constant
    self.show_progress=show_progress
    self.show_progress_nth_cycle=show_progress_nth_cycle
    self.evaluator = evaluator
    self.population_size = population_size
    self.f = f
    self.cr = cr
    self.n_cross = n_cross
    self.max_iter = max_iter
    self.monitor_cycle = monitor_cycle
    self.vector_length = evaluator.n
    self.eps = eps
    self.population = []
    self.seeded = False
    if insert_solution_vector is not None:
      assert len( insert_solution_vector )==self.vector_length
      self.seeded = insert_solution_vector
    for ii in xrange(self.population_size):
      self.population.append( np.zeros( (self.vector_length ) ) )
    
    # Fill the scores vector with arbitrary high values
    self.scores = np.zeros( (self.population_size) )
    self.scores.fill(1000)

    # Call optimizer
    self.optimize()
    
    # Get the best result back
    self.best_score = self.scores.min()
    self.best_vector = self.population[ self.scores.argmin() ]
    self.evaluator.x = self.best_vector
    
    # Final call to show_progress to report convergence to user. 
    if self.show_progress:
      self.evaluator.print_status(
        self.scores.min(),
        self.scores.mean(),
        self.population[ self.scores.argmin() ],
        'Final')


  def optimize(self):
    # initialise the population please
    self.make_random_population()
    # score the population please
    self.score_population()
    converged = False
    monitor_score = self.scores.min()
    self.count = 0
    while not converged:
      self.evolve()
      location = self.scores.argmin()
      if self.show_progress:
        if self.count%self.show_progress_nth_cycle==0:
          # make here a call to a custom print_status function in the evaluator function
          # the function signature should be (min_target, mean_target, best vector)
          self.evaluator.print_status(
            self.scores.min(),
            self.scores.mean(),
            self.population[ self.scores.argmin() ],
            self.count)

      self.count += 1
      
      # check for convergence

      ## First criteria: 
      #if self.count%self.monitor_cycle==0:
        ## Check if the minimum has decreased more than self.eps. Otherwise stop.
        #if (monitor_score - self.scores.min() ) < self.eps:
          #converged = True
        #else:
          #monitor_score = self.scores.min()

      ## Second criteria
      #rd = self.scores.mean() - self.scores.min() 
      #rd = rd*rd/self.scores.min()*self.scores.min() + self.eps 
      #if ( rd < self.eps ):
        #converged = True

      ## Third criteria
      #if self.count>=self.max_iter:
        #converged =True

      # Fourth criteria
      if self.scores.min() < self.eps:
        converged = True


  def make_random_population(self):
    for ii in xrange(self.vector_length):
      delta  = self.evaluator.domain[ii][1]-self.evaluator.domain[ii][0]
      offset = self.evaluator.domain[ii][0]
      random_values = np.random.random_sample(self.population_size)
      random_values = random_values*delta+offset
      # now please place these values ni the proper places in the
      # vectors of the population we generated
      for vector, item in zip(self.population, random_values):
        vector[ii] = item
    if self.seeded is not False:
      self.population[0] = self.seeded

  def score_population(self):
    vectors = []
    for vector in self.population:
      vectors.append(vector)
    self.scores = self.evaluator.target(vectors)

  def evolve(self):
    # Build collection of input vectors
    test_vectors = []
    for ii in xrange(self.population_size):
      permut = np.random.permutation(self.population_size - 1)
      # make parent indices (each i1, i2, i3 is a scalar)
      i1=permut[0]
      if (i1>=ii):
        i1+=1
      i2=permut[1]
      if (i2>=ii):
        i2+=1
      i3=permut[2]
      if (i3>=ii):
        i3+=1
      # pick out one population member with the scalar. 
      x1 = self.population[ i1 ]
      x2 = self.population[ i2 ]
      x3 = self.population[ i3 ]

      if self.f is None:
        use_f = np.random.random_sample()/2.0 + 0.5
      else:
        use_f = self.f

      vi = x1 + use_f*(x2-x3)
      # prepare the offspring vector please
      permut = np.random.permutation(self.vector_length)
      test_vector = self.population[ii].copy()
      # first the parameters that sure cross over
      for jj in xrange( self.vector_length  ):
        if (jj<self.n_cross):
          test_vector[ permut[jj] ] = vi[ permut[jj] ]
        else:
          if (rnd[jj]>self.cr):
            test_vector[ permut[jj] ] = vi[ permut[jj] ]
      test_vectors.append(test_vector) 

    # get the score please
    test_scores = self.evaluator.target( test_vectors )
    
    # check if the score is lower
    for ii in xrange(self.population_size):
      if test_scores[ii] < self.scores[ii] :
        self.scores[ii] = test_scores[ii]
        self.population[ii] = test_vectors[ii]


  def show_population(self):
    print "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
    for vec in self.population:
      print list(vec)
    print "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++"


class test_function(object):
  def __init__(self):
    self.x = None
    self.n = 9
    self.domain = [ (-100,100) ]*self.n
    self.optimizer =  differential_evolution_optimizer(self,population_size=100,n_cross=5)
    assert self.x*self.x.sum() < 1e-5

  def target(self, vector):
    tmp = vector.copy()
    result = ( np.cos(tmp * 10) + self.n + 1 ).sum() * ( (tmp)*(tmp) ).sum()
    return result


class test_rosenbrock_function(object):
  def __init__(self, dim=5):
    self.x = None
    self.n = 2*dim
    self.dim = dim
    self.domain = [ (-10,10) ]*self.n
    self.optimizer =  differential_evolution_optimizer(self, population_size = min(self.n*10,40), 
                                                       n_cross = self.n,
                                                       cr = 0.9, eps = 1e-12, 
                                                       show_progress = True)
    for x in self.x:
      assert abs(x-1.0)<1e-2


  def target(self, vectors):
    results = np.array([])
    for vector in vectors:
      # build parallel job. 
      tmp = vector.copy()
      x_vec = vector[0:self.dim]
      y_vec = vector[self.dim:]
      result=0
      for x,y in zip(x_vec,y_vec):
        # x is all the odds, y is all the evens
        # -> vector is a stack of self.dim odd entries and self.dim even entries
        result += 100.0 * ( (y - x * x ) ** 2.0 ) + ( 1 - x ) ** 2.0
      #print list(x_vec), list(y_vec), result
    # print(result)
      # For other problems. Need algorithm to analyze output and compute results accordingly. 
      # Run parallel job. Check when done. When done analyze result and append to results. 
      results = np.append(results, result)
    return results

  def print_status(self, mins,means,vector,txt):
    print txt,mins, means, list(vector)


def run():
  np.random.seed(0)
  test_rosenbrock_function(2)
  print "OK"


if __name__ == "__main__":
  run()


#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011-2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# stdlib imports
from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
from builtins import object
from abc import ABCMeta, abstractmethod

# GC3Pie imports
import gc3libs
from gc3libs.url import Url
from future.utils import with_metaclass


__docformat__ = 'reStructuredText'


class Store(with_metaclass(ABCMeta, object)):
    """
    Interface for storing and retrieving objects on permanent storage.

    Each `save` operation returns a unique "ID"; each ID is a Python
    string value, which is guaranteed to be temporally unique, i.e.,
    no two `save` operations in the same persistent store can result
    in the same IDs being assigned to different objects.  The "ID" is
    also stored in the instance attribute `_id`.

    Any Python object can stored, provided it meets the following
    conditions:

      * it can be pickled with Python's standard module `pickle`.
      * the instance attribute `persistent_id` is reserved for use by
        the `Store` class: it should not be set or altered by other
        parts of the code.
    """

    def __init__(self, url=None):
        if url and not isinstance(url, Url):
            url = Url(url)
        self.url = url

    def pre_fork(self):
        """
        Make preparations for `fork()`-ing the current process.

        This should close open network connections or any other
        sockets or file descriptors that cannot be used by both the
        parent and child process.

        The default implementation of this method does nothing; as of
        2018-04-10, the only subclass making use of this functionality
        is `SqlStore`:class:, which needs to dispose the SQLAlchemy
        engine and re-create it after forking.
        """
        pass

    def post_fork(self):
        """
        Restore functionality that was suspended in `pre_fork`:meth:

        This method will be called after forking/daemonizing has been
        successfully accomplished.

        The default implementation of this method does nothing.
        """
        pass

    def invalidate_cache(self):
        """
        Clear the loaded objects cache (if any).

        Subsequent `load()`:meth: calls are guaranteed to re-load the
        data directly from the backing store.
        """
        # The default implementation of this method does nothing, since
        # a `Store` subclass may not keep a cache of loaded objects.
        pass

    def list(self, **extra_args):
        """
        Return list of IDs of saved `Job` objects.

        This is an optional method; classes that do not implement it
        should raise a `NotImplementedError` exception.
        """
        raise NotImplementedError(
            "Method `list` not implemented in this class.")

    @abstractmethod
    def load(self, id_):
        """
        Load a saved object given its ID, and return it.
        """
        pass

    def _update_to_latest_schema(self):
        """
        Modify an object in-place to reflect changes in the schema.

        Called immediately after a successful `load()`, just before returning
        the retrieved object to the caller.
        """
        pass

    @abstractmethod
    def remove(self, id_):
        """
        Delete a given object from persistent storage, given its ID.
        """
        pass

    @abstractmethod
    def replace(self, id_, obj):
        """
        Replace the object already saved with the given ID with a copy
        of `obj`.
        """
        pass

    @abstractmethod
    def save(self, obj):
        """
        Save an object, and return an ID.
        """
        pass


# registration mechanism

_registered_store_ctors = {}


def register(scheme, constructor):
    """
    Register `constructor` as the factory corresponding to an URL scheme.

    If a different constructor is already registered for the same
    scheme, it is silently overwritten.

    The registry mapping schemes to constructors is used in the
    `make_store`:func: to create concrete instances of
    `gc3libs.persistence.Store`, given a URI that identifies the kind
    and location of the storage.

    :param str scheme: URL scheme to associate with the given constructor.

    :param callable constructor: A callable returning a `Store`:class:
      instance. Typically, a class constructor.
    """
    global _registered_store_ctors
    assert callable(constructor), (
        "Registering non-callable constructor for scheme "
        "'%s' in `gc3libs.persistence.register`"
        % scheme)
    gc3libs.log.debug(
        "Registering scheme '%s' with the `gc3libs.persistence` registry.",
        scheme)
    _registered_store_ctors[str(scheme)] = constructor


def make_store(uri, *args, **extra_args):
    """
    Factory producing concrete `Store`:class: instances.

    Given a URL and (optionally) initialization arguments, return a
    fully-constructed `Store`:class: instance.

    The only required argument is `uri`; if any other arguments are
    present in the function invocation, they are passed verbatim to
    the constructor associated with the scheme of the given `uri`.

    Example::

      >>> fs1 = make_store('file:///tmp')
      >>> fs1.__class__.__name__
      'FilesystemStore'

    Argument `uri` can also consist of a path name, in which case a
    URL scheme 'file:///' is assumed::

      >>> fs2 = make_store('/tmp')
      >>> fs2.__class__.__name__
      'FilesystemStore'

    """
    if not isinstance(uri, Url):
        uri = Url(uri)

    # since SQLAlchemy allows URIs of the form `db+driver://...`
    # (e.g., `postresql+psycopg://...`) we need to examine the URI
    # scheme only up to the first `+`
    scheme = uri.scheme.split('+')[0]

    try:
        # hard-code schemes that are supported by GC3Pie itself
        if uri.scheme == 'file':
            import gc3libs.persistence.filesystem
            return gc3libs.persistence.filesystem.make_filesystemstore(
                uri, *args, **extra_args)
        elif scheme in [
                # DBs supported in SQLAlchemy core as of version 1.1,
                # see: http://docs.sqlalchemy.org/en/latest/dialects/index.html
                'firebird',
                'mssql',
                'mysql',
                'oracle',
                'postgresql',
                'sqlite',
                'sybase',
        ]:
            import gc3libs.persistence.sql
            return gc3libs.persistence.sql.make_sqlstore(
                uri, *args, **extra_args)
        else:
            try:
                return _registered_store_ctors[
                    uri.scheme](uri, *args, **extra_args)
            except KeyError:
                gc3libs.log.error(
                    "Unknown URL scheme '%s' in"
                    " `gc3libs.persistence.make_store`:"
                    " has never been registered.", uri.scheme)
                raise
    except Exception as err:
        gc3libs.log.error(
            "Error constructing store for URL '%s': %s: %s",
            uri, err.__class__.__name__, err)
        raise

# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="store",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   gbraunian.py -- Front-end script for evaluating Matlab function
#   `braunian` over large number of events.
#
#   Copyright (C) 2016, 2017  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `Matlab` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gbraunian.py --help`` for program usage
instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2016-07-14:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import gbraunian
    gbraunian.GbraunianScript().run()

import os
import sys
import time
import tempfile

import tarfile
import shutil
import pandas
import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_REMOTE_OUTPUT_FILE = "workflow.mat"
DEFAULT_FUNCTION = "brownian_cloud"
DEFAULT_CASE_FILE="3d_case1.txt"
DEFAULT_CHUNK=1
MATLAB_CMD="matlab -nosplash -nodisplay -nodesktop -r \"{main_function}({events},'{case_file}','{output_file}');quit\""

## custom application class
class GbraunianApplication(Application):
    """
    Custom class to wrap the execution of the Matlab function
    over a subset of the total number of events.
    """
    application_name = 'gbraunian'

    def __init__(self, events, matlab_file, case_file, **extra_args):

        executables = []
        inputs = dict()
        outputs = dict()

        inputs[matlab_file] = os.path.basename(matlab_file)
        matlab_function = inputs[matlab_file].split('.')[0]
        inputs[case_file] = os.path.basename(case_file)

        arguments = MATLAB_CMD.format(main_function=matlab_function,
                                      events=events,
                                      case_file=os.path.basename(case_file),
                                      output_file=DEFAULT_REMOTE_OUTPUT_FILE)

        # Set output
        outputs[DEFAULT_REMOTE_OUTPUT_FILE] = DEFAULT_REMOTE_OUTPUT_FILE

        gc3libs.log.debug("Creating application for executing: %s",
                          arguments)

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gbraunian.log',
            join=True,
            executables = executables,
            **extra_args)


class GbraunianScript(SessionBasedScript):
    """
    Take total number of events and create a list of chunked events.
    For each chunk, run the provided Matlab function.

    The ``gbraunian`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gbraunian`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gbraunian``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GbraunianApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GbraunianApplication,
            )

    def setup_args(self):

        self.add_param('events', type=int,
                       help="Total number of events.")

        self.add_param('main',
                       help="Matlab function to execute.")

        self.add_param('case',
                       help="case file.")

    def setup_options(self):
        self.add_param("-k", "--chunk", metavar="INT", type=int,
                       dest="chunk", default=DEFAULT_CHUNK,
                       help="How to split the edges input data set. "
                       "Default: %(default)s.")

    def parse_args(self):
        """
        Check for validity of input arguments
        """
        try:
            assert isinstance(self.params.events,int), \
                "Invalid number of events '%s'. Must be positive integer." % str(self.params.events)

            assert os.path.isfile(self.params.case), \
                "case file '%s' not found" % self.params.case

            assert os.path.isfile(self.params.main), \
                "Matlab function file '%s' not found" % self.params.main

            gc3libs.log.info("Using matlab function name: '%s'" % os.path.basename(self.params.main).split('.')[0])

        except AssertionError as ex:
            raise ValueError(ex.message)

    def new_tasks(self, extra):
        """
        Read content of 'command_file'
        For each command line, generate a new Application
        """
        tasks = []

        for event in get_events(self.params.events,self.params.chunk):
            extra_args = extra.copy()
            tasks.append(GbraunianApplication(
                event,
                self.params.main,
                self.params.case,
                **extra_args))

        return tasks

def get_events(events,chunk):
    event_list = [chunk for elem in range(1,events/chunk+1)]
    if events % chunk:
        event_list.append(events % chunk)
    return event_list



#! /usr/bin/env python
#
#   gfsurfer.py -- Front-end script for running the docking program rDock
#   over a list of ligand files.
#
#   Copyright (C) 2014, 2015  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gfsurfer.py --help`` for program usage instructions.

Input parameters consists of:
...

Options:
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2015-02-17:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gfsurfer
    gfsurfer.GfsurferScript().run()

import os
import sys
import time
import tempfile
import re

import shutil
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_CORES = 1
DEFAULT_MEMORY = Memory(3000,MB)

DEFAULT_REMOTE_OUTPUT_FOLDER="./output/"

INPUT_LIST_PATTERNS = [".nii",".nii.tgz",".nii.gz",".nii.tar.gz"]
FREESURFER_STEPS = ['cross','long']

## custom application class
class GfsurferApplication(Application):
    """
    """
    application_name = 'gfsurfer'

    def __init__(self, subject_name, input_nifti, freesurfer_steps, **extra_args):

        output_dir = DEFAULT_REMOTE_OUTPUT_FOLDER + subject_name + ".crossTP1"

        inputs = dict()
        outputs = dict()

        gfsurfer_wrapper = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gfsurfer_wrapper.py")
        inputs[gfsurfer_wrapper] = os.path.basename(gfsurfer_wrapper)
        inputs[input_nifti] = os.path.basename(input_nifti)
        outputs[output_dir] = output_dir

        arguments = "./%s %s %s %s" % (inputs[gfsurfer_wrapper],
                                       subject_name,
                                       os.path.basename(input_nifti),
                                       DEFAULT_REMOTE_OUTPUT_FOLDER)

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gfsurfer.log',
            join=True,
            executables = [os.path.basename(gfsurfer_wrapper)],
            **extra_args)

class GfsurferScript(SessionBasedScript):
    """

    The ``gfsurfer`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gfsurfer`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GfsurferApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GfsurferApplication,
            )

    def setup_args(self):
        self.add_param('input_data', type=str,
                       help="Root localtion of input data. "
                       "Note: expected folder structure: "
                       " - input files in case of cross sectional runs. "
                       " - subdirectories with input files for more TP in case of longitudinal ")

        self.add_param("-P", "--pipeline", metavar="STRING", type=str,
                       dest="pipeline",
                       default="cross",
                       help="Comma separated list of Freesurfer steps to be "
                       " executed on each valid input file. "
                       " Valid values: %s." % FREESURFER_STEPS)

    def parse_args(self):
        try:
            assert os.path.isdir(self.params.input_data), "Input folder %s not found" % self.params.input_data
        except AssertionError as ex:
            raise OSError(ex.message)

        self.freesurfer_pipeline = [ step for step in self.params.pipeline if step in FREESURFER_STEPS ]
        gc3libs.log.debug("Accepted following Freesurfer steps: %s" % self.freesurfer_pipeline)

    def new_tasks(self, extra):
        """
        For each input folder, create an instance of GfsurferApplication
        """
        tasks = []

        for input_nifti in [ nifti for nifti in os.listdir(self.params.input_data) if
                             any(nifti.endswith(pattern) for pattern in INPUT_LIST_PATTERNS) ]:

            # filename example: 0103645_anat.nii.gz
            # extract root folder name to be used as jobname
            subject_name = input_nifti.split(".")[0]

            extra_args = extra.copy()
            extra_args['jobname'] = subject_name

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                        'run_%s' % subject_name)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                        'run_%s' % subject_name)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                        'run_%s' % subject_name)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                        'run_%s' % subject_name)

            tasks.append(GfsurferApplication(
                subject_name,
                os.path.join(self.params.input_data,input_nifti),
                self.freesurfer_pipeline,
                **extra_args))

        return tasks



#! /usr/bin/env python
#
#   gdocking.py -- Front-end script for submitting ROSETTA `docking_protocol` jobs to SMSCG.
#
#   Copyright (C) 2010-2012  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting ROSETTA jobs to SMSCG.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gdocking --help`` for program usage instructions.
"""

from __future__ import absolute_import, print_function

__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
# summary of user-visible changes
__changelog__ = """
  2012-02-28:
    * Use option ``-R`` to request a specific release of Rosetta.
  2011-05-06:
    * Workaround for Issue 95: now we have complete interoperability
      with GC3Utils.
  2011-02-10:
    * Renamed option '-t'/'--tarfile' to '-T'/'--collect',
      in order not to conflict with `SessionBasedScript`
      option '-t'/'--table'.
    * Removed the '-b' option.
  2010-12-20:
    * Renamed to ``gdocking``; default session file is not ``gdocking.csv``.
  2010-09-21:
    * Do not collect output FASC and PDB files into a single tar file.
      (Get old behavior back with the '-t' option)
    * Do not compress PDB files in Rosetta output.
      (Get the old behavior back with the '-z' option)
  2010-08-09:
    * Exitcode tracks job status; use the "-b" option to get the old behavior back.
      The new exitcode is a bitfield; the 4 least-significant bits have the following
      meaning:
         ===    ============================================================
         Bit    Meaning
         ===    ============================================================
         0      Set if a fatal error occurred: `grosetta` could not complete
         1      Set if there are jobs in `FAILED` state
         2      Set if there are jobs in `RUNNING` or `SUBMITTED` state
         3      Set if there are jobs in `NEW` state
         ===    ============================================================
      This boils down to the following rules:
         * exitcode == 0: all jobs are `DONE`, no further `grosetta` action
         * exitcode == 1: an error interrupted `grosetta` execution
         * exitcode == 2: all jobs finished, but some are in `FAILED` state
         * exitcode > 3: run `grosetta` again to progress jobs
    * when all jobs are finished, exit `grosetta` even if the "-C" option is given
    * Print only summary of job statuses; use the "-l" option to get the long listing
  2010-07-26:
    * Default output directory is now './' (should be less surprising to users).
    * FASC and PDBs are now collected in the output directory.
  2010-07-15:
    * After successful retrieval of job information, reorder output files so that:
      - for each sumitted job, there is a corresponding ``input.N--M.fasc`` file,
        in the same directory as the input ".pdb" file;
      - all decoys belonging to the same input ".pdb" file are collected into
        a single ``input.decoys.tar`` file (in the same dir as the input ".pdb" file);
      - output from grid jobs is kept untouched in the "job.XXX/" directories.
    * Compress PDB files by default, and prefix them with a "source filename + N--M" prefix
    * Number of computed decoys can now be increased from the command line:
      if `grosetta` is called with different '-P' and '-p' options, it will
      add new jobs to the list so that the total number of decoys per input file
      (including already-submitted ones) is up to the new total.
    * New '-N' command-line option to discard old session contents and start a new session afresh.
  2010-07-14:
    * Default session file is now './grosetta.csv', so it's not hidden to users.
"""
__docformat__ = 'reStructuredText'


# workaround for Issue 95,
# see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gdocking
    gdocking.GDockingScript().run()


# stdlib imports
import grp
import logging
import os
import os.path
import pwd
import sys
import tarfile
import time

# interface to Gc3libs
import gc3libs
from gc3libs.application.rosetta import RosettaDockingApplication
from gc3libs.cmdline import SessionBasedScript, existing_file, positive_int


## The GDocking application

class GDockingApplication(RosettaDockingApplication):
    """
    Augment a `RosettaDockingApplication` with state transition
    methods that implement job status reporting for the UI, and data
    post-processing.
    """
    def __init__(self, pdb_file_path, native_file_path=None,
                 number_of_decoys_to_create=1, flags_file=None,
                 collect=False, **extra_args):
        RosettaDockingApplication.__init__(
            self, pdb_file_path, native_file_path,
            number_of_decoys_to_create, flags_file,
            **extra_args)
        # save pdb_file_path for later processing
        self.pdb_file_path = pdb_file_path
        # define additional attributes
        self.collect = collect, # whether to collect result PDBs into a tarfile
        self.computed = 0 # number of decoys actually computed by this job

    def terminated(self):
        output_dir = self.output_dir
        # work directory is the parent of the download directory
        work_dir = os.path.dirname(output_dir)
        # move around output files so they're easier to preprocess:
        #   1. All '.fasc' files land in the same directory as the input '.pdb' file
        #   2. All generated '.pdb'/'.pdb.gz' files are collected in a '.decoys.tar'
        #   3. Anything else is left as-is
        input_name = os.path.basename(self.pdb_file_path)
        input_name_sans = os.path.splitext(input_name)[0]
        output_tar_filename = os.path.join(output_dir, 'docking_protocol.tar.gz')
        # count: 'protocols.jobdist.main: Finished 1brs.0--1.1brs_0002 in 149 seconds.'
        if os.path.exists(output_tar_filename):
            output_tar = tarfile.open(output_tar_filename, 'r:gz')
            # single tar file holding all decoy .PDB files
            pdbs_tarfile_path = os.path.join(work_dir, input_name_sans) + '.decoys.tar'
            if self.collect:
                if not os.path.exists(pdbs_tarfile_path):
                    pdbs = tarfile.open(pdbs_tarfile_path, 'w')
                else:
                    pdbs = tarfile.open(pdbs_tarfile_path, 'a')
            for entry in output_tar:
                if (entry.name.endswith('.fasc') or entry.name.endswith('.sc')):
                    filename, extension = os.path.splitext(entry.name)
                    scoring_file_name = (os.path.join(work_dir, input_name_sans)
                                         + '.' + self.jobname + extension)
                    src = output_tar.extractfile(entry)
                    dst = open(scoring_file_name, 'wb')
                    dst.write(src.read())
                    dst.close()
                    src.close()
                elif (self.collect and
                      (entry.name.endswith('.pdb.gz') or entry.name.endswith('.pdb'))):
                    src = output_tar.extractfile(entry)
                    dst = tarfile.TarInfo(entry.name)
                    dst.size = entry.size
                    dst.type = entry.type
                    dst.mode = entry.mode
                    dst.mtime = entry.mtime
                    dst.uid = os.getuid()
                    dst.gid = os.getgid()
                    dst.uname = pwd.getpwuid(os.getuid()).pw_name
                    dst.gname = grp.getgrgid(os.getgid()).gr_name
                    if hasattr(entry, 'pax_headers'):
                        dst.pax_headers = entry.pax_headers
                    pdbs.addfile(dst, src)
                    src.close()
            if self.collect:
                pdbs.close()
        else: # no `docking_protocol.tar.gz` file
            self.info = ("No 'docking_protocol.tar.gz' file found.")


## the script class

class GDockingScript(SessionBasedScript):
    """
Compute decoys of specified '.pdb' files by running several
Rosetta 'docking_protocol' instances in parallel.

The `gdocking` command keeps a record of jobs (submitted, executed and
pending) in a session file; at each invocation of the command, the
status of all recorded jobs is updated, output from finished jobs is
collected, and a summary table of all known jobs is printed.

If any INPUT argument is specified on the command line, `gdocking`
appends new jobs to the session file, up to the quantity needed
to compute the requested number of decoys.  Each of the INPUT
parameters can be either a single '.pdb' file, or a directory, which
is recursively scanned for '.pdb' files.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; `gdocking` will delay submission
of newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            # Use fully-qualified name for class,
            # to allow GC3Utils to unpickle job instances
            # see: https://github.com/uzh/gc3pie/issues/95
            application = GDockingApplication,
            input_filename_pattern = '*.pdb'
            )

    def setup_options(self):
        self.add_param("-f", "--flags-file", dest="flags_file",
                       default=os.path.join(gc3libs.Default.RCDIR, 'docking_protocol.flags'),
                       metavar="PATH",
                       help="Pass the specified flags file to Rosetta 'docking_protocol'"
                       " Default: '%(default)s'"
                       )
        self.add_param("-P", "--decoys-per-file", dest="decoys_per_file",
                       type=positive_int, default=1,
                       metavar="NUM",
                       help="Compute NUM decoys per input file (default: %(default)s)."
                       )
        self.add_param("-p", "--decoys-per-job", dest="decoys_per_job",
                       type=positive_int, default=1,
                       metavar="NUM",
                       help="Compute NUM decoys in a single job (default: %(default)s)."
                       " This parameter should be tuned so that the running time"
                       " of a single job does not exceed the maximum wall-clock time."
                       )
        self.add_param("-R", "--release",
                       type=str, dest="rosetta_release", default="3.1",
                       metavar="NAME",
                       help="Numerical suffix to identify which version of Rosetta should be requested."
                       " For example: '-e 20110622' will run rosetta-svn20110622."
                       " (default: %(default)s)"
                       )
        self.add_param("-T", "--collect", dest="collect", default=False, action="store_true",
                       help="Collect all output PDB and FASC/SC files into a single '.tar' file,"
                       " located in the output directory (see the '-o' option)."
                       )
        self.add_param("-z", "--compress-pdb", dest="compress", default=False, action="store_true",
                       help="Compress '.pdb' output files with `gzip`."
                       )

    def parse_args(self):
        self.instances_per_file = self.params.decoys_per_file
        self.instances_per_job = self.params.decoys_per_job

        self.extra['application_release'] = self.params.rosetta_release
        self.extra['number_of_decoys_to_create'] = self.params.decoys_per_job
        self.extra['collect'] = self.params.collect

        try:
            existing_file(self.params.flags_file)
        except Exception, ex:
            gc3libs.log.error("Invalid flags file `%s`: please supply a valid flags file by adding the option `-f PATH`" % self.params.flags_file)
            raise ex

        self.extra['flags_file'] = os.path.abspath(self.params.flags_file)
        self.log.info("Using flags file '%s'", self.extra['flags_file'])



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2012-2013,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'


import gc3libs
from gc3libs.cmdline import SessionBasedDaemon
import os
from gc3libs.quantity import GiB, MiB

import inotifyx

class GBeastApp(gc3libs.Application):
    def __init__(self, beast, jarfile, path, ncores, **extra):
        self.fname = os.path.basename(path)
        self.infile = path
        self.extra = extra
        extra['output_dir'] = self.fname[:-4] + '.d'
        extra['requested_cores'] = ncores
        extra['requested_memory'] = ncores*4000*MiB
        extra['jobname'] = "GBeastApp.%s" % self.fname
        args = ['java',
                '-Xmx%dm' % extra['requested_memory'].amount(MiB),
                '-jar', jarfile,
                '-threads', ncores,
                '-beagle_instances', ncores,
                self.fname]

        gc3libs.Application.__init__(
            self,
            arguments = args,
            inputs = [self.infile],
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'stdout.txt',
            stderr = 'stderr.txt',
            **extra)

class GBeastDaemon(SessionBasedDaemon):
    """This daemon script will continuosly check one or more `inbox`
    directories and, for each *new* XML file created in it, will create
    a new job running BEAST on it.

    By default it will run BEAST v1.8, but if the filename contains
    the string 'BEAST2' it will use BEAST v2 instead. Path to the
    correct Beast jar can be specified throught option `--beast1` and
    `--beast2`.

    """
    version = '0.1'

    def setup_options(self):
        self.add_param('--beast1', default='/apps/BEASTv1.8.2/lib/beast.jar', help='Path to BEAST v1 jar file')
        self.add_param('--beast2', default='/apps/BEASTv2.3.2/lib/beast.jar', help='Path to BEAST v2 jar file')
        self.add_param('--cores', default=1, type=int, help="Amount of cores to use. Default: %(default)s")


    def new_tasks(self, extra, epath=None, emask=0):
        # Scan of the directory is performed in `before_main_loop`,
        # here we only need to ensure the fake always-running task is
        # created.

        if not epath:
            # The script has just started, don't do anything yet, as
            # we will always wait for inotify for new files.
        elif epath.endswith('.xml'):
            gc3libs.log.info("Ignoring file `%s` since it doesn't end in .xml" % epath)
            return []

        beast = 'beast1'
        if 'BEAST2' in epath:
            beast = 'beast2'
        else:
            gc3libs.log.warning("Unable to guess which version of BEAST you"
                                " want to run for file %s. Assuming BEAST v1" % epath)

        jarfile = self.params.beast1 if beast == 'beast1' else self.params.beast2
        try:
            # We need to load from a previously saved job extra
            # arguments, otherwise we don't know where to get
            # them.

            # Check if the file is being processed already
            if epath in [app.path for app in self.session if 'path' in app]:
                gc3libs.log.warning(
                    "Ignoring file `%s`, as it's been processed already."
                    " You can re-process it by renaming the file or using"
                    " `gresub`" % epath)
                return []

            if emask == inotifyx.IN_CLOSE_WRITE:
                app = GBeastApp(beast,
                                jarfile,
                                epath,
                                self.params.cores,
                                **extra.copy())
                gc3libs.log.info("Added new application to session for file %s", epath)
            else:
                gc3libs.log.debug("Ignoring notify mask `%s` for file `%s`" % (emask, epath))
            return [app]
        except Exception as ex:
            gc3libs.log.error("Error while adding application for file %s: %s", epath, ex)
        return []

if "__main__" == __name__:
    from gbeastdaemon import GBeastDaemon
    GBeastDaemon().run()
